//================================================================================================
//                                  cIotModuleSte header
//================================================================================================

class cIotModuleApp;

class cIotModuleSte
{
  public:
  
    static const uint8_t MAX_CHANNELS = 8;
    static const uint8_t PIN_STEP = 0;
    static const uint8_t PIN_DIR = 1;
    
  private:
      
#if IOTMODULE_STEPPER_TIMING == 8
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_8_BIT;
    static const uint32_t START_SPEED = 320; 
    static const uint32_t MAX_SPEED = 307200;
#elif IOTMODULE_STEPPER_TIMING == 7
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_9_BIT;
    static const uint32_t START_SPEED = 160;
    static const uint32_t MAX_SPEED = 153600;
#elif IOTMODULE_STEPPER_TIMING == 6
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_10_BIT;
    static const uint32_t START_SPEED = 80;
    static const uint32_t MAX_SPEED = 76800;
#elif IOTMODULE_STEPPER_TIMING == 5
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_11_BIT;
    static const uint32_t START_SPEED = 40;
    static const uint32_t MAX_SPEED = 38400;
#elif IOTMODULE_STEPPER_TIMING == 4
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_12_BIT;
    static const uint32_t START_SPEED = 20;
    static const uint32_t MAX_SPEED = 19200;
#elif IOTMODULE_STEPPER_TIMING == 3
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_13_BIT;
    static const uint32_t START_SPEED = 10;
    static const uint32_t MAX_SPEED = 9600;
#elif IOTMODULE_STEPPER_TIMING == 2
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_14_BIT;
    static const uint32_t START_SPEED = 5;
    static const uint32_t MAX_SPEED = 4800;
#elif IOTMODULE_STEPPER_TIMING == 1
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_15_BIT;
    static const uint32_t START_SPEED = 3;
    static const uint32_t MAX_SPEED = 2400;
#elif IOTMODULE_STEPPER_TIMING == 0
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_16_BIT;
    static const uint32_t START_SPEED = 2;
    static const uint32_t MAX_SPEED = 1200;
#endif

    enum DIR : int8_t
    {
      DIR_POS = 1,
      DIR_NEG = -1,
    };
    
    enum STATUS : int8_t
    {
      RUN = 2,
      RUN_IDLE = 1,
      IDLE = 0,
      IDLE_RUN = -1,
    };
    
    struct sSingleParams
    {
      bool Attached;
      gpio_num_t PinStepName;
      gpio_num_t PinDirName;
      gpio_pull_mode_t PinStepResetMode;
      gpio_pull_mode_t PinDirResetMode;
      uint32_t SpeedHz;
      uint16_t AccHz_mSec;
      int64_t ActualPosition;
      int64_t ActualDestination;
      int32_t PreviousSpeed;
      DIR PreviousDirection;
      STATUS PreviousStatus;
      bool PreviousClockSignalEnabled;
      cExtend_16_64 Extend_16_64;
      bool ResetActive;
      
    };

    struct sAllData
    {
      struct sSingleParams Params[MAX_CHANNELS];
    };

    struct sAllData Data;
    
    cIotModuleApp *pIotModuleApp;
    float AverageLoopTimeMS;
    int64_t OldUS;
    ledc_timer_config_t ledc_timer;
    ledc_channel_config_t ledc_channel;
    pcnt_config_t pcnt_unit;
	
	uint8_t test;
    
  public:
  
    void Initialize(cIotModuleApp *pIotModuleApp);
    uint32_t GetDistanceFromDestination(const uint8_t Channel);
    void SetDestination(const uint8_t Channel, const uint32_t Destination);
    void Stop(const uint8_t Channel);
    void SetControlPin(const uint8_t Channel, const uint8_t Type, const gpio_num_t Name, const gpio_pull_mode_t ResetMode);
    bool Attach(uint8_t LedcTimer, uint8_t LedcChannel, uint8_t PcntChannel);
    bool Detach(const uint8_t Channel);
    bool ConfigSpeedAcc(const uint8_t Channel, const uint32_t SpeedHz, const uint16_t AccHz_mSec);
    void TimedInterpolate();
    
  private:
  
    uint32_t AddOffset(const uint32_t Value);
    void SetCounterDirection(const uint8_t Channel, const DIR Direction);
    void SetSpeed(const uint8_t Channel, const uint32_t Speed);
};
