//================================================================================================
//                                  cIotModuleApp body
//================================================================================================

void cIotModuleApp::Initialize(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);

  char pName[32 + 1];
  GetModuleName(pName, sizeof(pName));
  //  adc calibration
  if (esp_adc_cal_check_efuse(ESP_ADC_CAL_VAL_EFUSE_TP) == ESP_OK)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration - eFuse two point supported", __FILENAME__, __LINE__);
  }
  else if (esp_adc_cal_check_efuse(ESP_ADC_CAL_VAL_EFUSE_VREF) == ESP_OK)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration - eFuse Vref supported", __FILENAME__, __LINE__);
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration not supported", __FILENAME__, __LINE__);
  }
  //  set attenuation
  analogSetAttenuation(ADC_11db);
  //  Characterize ADC at particular attenuation
  esp_adc_cal_value_t val_type = esp_adc_cal_characterize(ADC_UNIT_1, ADC_ATTEN_DB_11, ADC_WIDTH_BIT_12, 0, &AdcChars);
  //  Check type of calibration value used to characterize ADC
  if (val_type == ESP_ADC_CAL_VAL_EFUSE_TP)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration done with eFuse two point", __FILENAME__, __LINE__);
  }
  else if (val_type == ESP_ADC_CAL_VAL_EFUSE_VREF)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ADC calibration done with eFuse Vref", __FILENAME__, __LINE__);
  }

  //  adc1_config_width(ADC_WIDTH_BIT_12);
  int val = hall_sensor_read();
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Hall sensor value : %d", __FILENAME__, __LINE__, val);
  
  //  initialize ISR service for GPIO
  ESP_ERROR_CHECK(gpio_install_isr_service(ESP_INTR_FLAG_LEVEL1));
  //  uninstall is never used
  //  gpio_uninstall_isr_service();

  IotModuleSWEncoder.Initialize(this);
  IotModuleHWEncoder.Initialize(this);
  IotModulePeriod.Initialize(this);
  IotModuleAdc.Initialize(this);
  IotModuleSte.Initialize(this);
  IotModuleCounter.Initialize(this);
  IotModuleServo.Initialize(this);
  IotModulePWM.Initialize(this);

  //  start count total active pins
  AllPinsData.TotalPinsNumber = 0;

  //  real pins number
  uint8_t InOutPins = IOTMODULE_REAL_PINS;
  //  if less pins wanted
  if (IOTMODULE_INOUT_PINS < InOutPins)
  {
    //  less pins used
    InOutPins = IOTMODULE_INOUT_PINS;
  }
  AllPinsData.TotalPinsNumber += InOutPins;

  //  if at least one adc channel is defined
  if (IOTMODULE_ADC24_PINS > 0)
  {
    //  set AD7124 control pins
    for (uint8_t i = 0; (i < IOTMODULE_REAL_PINS); i++)
    {
      //  if this pin is reserved for adc24
      if ( ModuleProps.PinsProps[i].Name == AD7124_PIN_NUM_MOSI ||
           ModuleProps.PinsProps[i].Name == AD7124_PIN_NUM_MISO ||
           ModuleProps.PinsProps[i].Name == AD7124_PIN_NUM_CLK)
      {
        //  set ADC24 GPIO parameters
        switch(ModuleProps.PinsProps[i].Name)
        {
          case AD7124_PIN_NUM_MOSI:
            IotModuleAdc.SetSPI_Control_Pin(cIotModuleAdc::AD7124_MOSI, (gpio_num_t)ModuleProps.PinsProps[i].Name, ModuleProps.PinsProps[i].ResetMode);
            break;
          case AD7124_PIN_NUM_MISO:
            IotModuleAdc.SetSPI_Control_Pin(cIotModuleAdc::AD7124_MISO, (gpio_num_t)ModuleProps.PinsProps[i].Name, ModuleProps.PinsProps[i].ResetMode);
            break;
          case AD7124_PIN_NUM_CLK:
            IotModuleAdc.SetSPI_Control_Pin(cIotModuleAdc::AD7124_CLK, (gpio_num_t)ModuleProps.PinsProps[i].Name, ModuleProps.PinsProps[i].ResetMode);
            break;
        }
      }
    }
    uint8_t ADC24VirtPinName = FIRST_AD7124_PIN_NAME;
    uint8_t ADC24VirtPinCount = IOTMODULE_ADC24_PIN_MASTER + IOTMODULE_ADC24_PINS;
    //  adc24 master pin
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Name = ADC24VirtPinName++;
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Channel1 = ADC24_MasterPinProps.Channel1;
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].CapName = ADC24_MasterPinProps.CapName;
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].ResetMode = ADC24_MasterPinProps.ResetMode;
    ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].TypesSize = ADC24_MasterPinProps.TypesSize;
    //  copy the pin types
    for (uint8_t i = 0; i < ADC24_MasterPinProps.TypesSize; i++)
    {
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Types[i] = ADC24_MasterPinProps.Types[i];
    }
    ADC24VirtPinCount--;
    AllPinsData.TotalPinsNumber++;
    //  adc24 data pins
    for (; ADC24VirtPinCount != 0; ADC24VirtPinCount--)
    {
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Name = ADC24VirtPinName++;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Channel1 = ADC24ChannelPinProps.Channel1;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].CapName = ADC24ChannelPinProps.CapName;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].ResetMode = ADC24ChannelPinProps.ResetMode;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].TypesSize = ADC24ChannelPinProps.TypesSize;
      //  copy the pin types
      for (uint8_t i = 0; i < ADC24ChannelPinProps.TypesSize; i++)
      {
        ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Types[i] = ADC24ChannelPinProps.Types[i];
      }
      AllPinsData.TotalPinsNumber++;
    }
  }

  //  if at least one generic pin is defined
  if (IOTMODULE_GENERIC_PINS > 0)
  {
    //  compile the props for the generic pins defined
    uint8_t GenVirtPinName = FIRST_GENERIC_PIN_NAME;
    uint8_t GenVirtPinCount = IOTMODULE_GENERIC_PINS;
    for (; GenVirtPinCount != 0; GenVirtPinCount--)
    {
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Name = GenVirtPinName++;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Channel1 = GenericPinProps.Channel1;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].CapName = GenericPinProps.CapName;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].ResetMode = GenericPinProps.ResetMode;
      ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].TypesSize = GenericPinProps.TypesSize;
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TypesSize : %d", __FILENAME__, __LINE__, GenericPinProps.TypesSize);
      //  copy the pin types
      for (uint8_t i = 0; i < GenericPinProps.TypesSize; i++)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Types : %d", __FILENAME__, __LINE__, GenericPinProps.Types[i]);
        ModuleProps.PinsProps[AllPinsData.TotalPinsNumber].Types[i] = GenericPinProps.Types[i];
      }
      AllPinsData.TotalPinsNumber++;
    }
  }

  //  reset floating values and pin type
  for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
  {
    AllPinsData.PinsParams[Index].fValue = 0.0f;
    AllPinsData.PinsParams[Index].Type = UNUSED;
  }
  //  initialize pin->index lookup table
  for (uint8_t Index = 0; Index < IOTMODULE_LOOKUP_TABLE_SIZE; Index++)
  {
    //  set all index values to 255 == not valid
    PinIndexLookupTable[Index] = 255;
  }
  //  build pin->index lookup table
  for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
  {
    PinIndexLookupTable[ModuleProps.PinsProps[Index].Name] = Index;
  }

  //  module is not programmed
  ProgStatus = NOTHING;

  //  display pinout type
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_PINOUT_TYPE : %d", __FILENAME__, __LINE__, IOTMODULE_PINOUT_TYPE);
  //  display I/O module pins count
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_INOUT_PINS : %d", __FILENAME__, __LINE__, IOTMODULE_INOUT_PINS);
  //  display adc24 pins count
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_ADC24_PINS : %d", __FILENAME__, __LINE__,  IOTMODULE_ADC24_PIN_MASTER + IOTMODULE_ADC24_PINS);
  //  display generic pins count
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_GENERIC_PINS : %d", __FILENAME__, __LINE__, IOTMODULE_GENERIC_PINS);
  //  display total handled pins count
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - internal TotalPinsNumber : %d", __FILENAME__, __LINE__, AllPinsData.TotalPinsNumber);
  //  display pin list
  for (uint8_t i = 0; i < AllPinsData.TotalPinsNumber; i++)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Pin List , Index : %d , Name : %d, TypesSize : %d", __FILENAME__, __LINE__, i, ModuleProps.PinsProps[i].Name, ModuleProps.PinsProps[i].TypesSize);
    for (uint8_t j = 0; j < ModuleProps.PinsProps[i].TypesSize; j++)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Pin List , Type : %d", __FILENAME__, __LINE__, ModuleProps.PinsProps[i].Types[j]);
    }
  }
}

bool cIotModuleApp::WiFiModeOverridePinDefined()
{
  bool Found = false;
  if (IOTMODULE_WIFI_MODE_PIN != -1)
  {
    //  for every physical pins
    for (uint8_t i = 0; (i < IOTMODULE_REAL_PINS); i++)
    {
      //  if the pin is one of the physical pins
      if (ModuleProps.PinsProps[i].Name == (uint8_t)IOTMODULE_WIFI_MODE_PIN)
      {
        //  found
        Found = true;
        break;
      }
    }
    if (!Found)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - configured IOTMODULE_WIFI_MODE_PIN %d : Pin number is invalid", __FILENAME__, __LINE__, IOTMODULE_WIFI_MODE_PIN);
    }
  }
  return Found;
}

bool cIotModuleApp::WiFiModeOverridePinStatus()
{
  //  return status of the pin
  bool Level = false;
  if (IOTMODULE_WIFI_MODE_PIN != -1)
  {
    Level = gpio_get_level((gpio_num_t)IOTMODULE_WIFI_MODE_PIN) == 0 ? false : true;
  }
  return Level;
}

void cIotModuleApp::ADC24SetParam(const uint16_t Sps, const bool MultipleSetups, const bool FiltersSingleCycle, const uint8_t PreFilterType,
                                  const uint8_t PostFilterType, const uint16_t ActiveChannels, const uint16_t BiasedChannels, const uint16_t DiffChannels,
                                  const uint16_t PseudoChannels, const uint8_t ChannelsGains0, const uint8_t ChannelsGains1, const uint8_t ChannelsGains2)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - ADC24SetParam(App) called", __FILENAME__, __LINE__);
  //  call IotModuleAdc.SetParam
  IotModuleAdc.SetParam(Sps, MultipleSetups, FiltersSingleCycle, PreFilterType,
                        PostFilterType, ActiveChannels, BiasedChannels, DiffChannels,
                        PseudoChannels, ChannelsGains0, ChannelsGains1, ChannelsGains2);
  //  set AD7124 section as programmed
  ProgStatus = static_cast<PROG_STATUS>(ProgStatus | AD7124);
}

bool cIotModuleApp::StepperConfigSpeedAcc(uint8_t PinIndex, uint32_t SpeedHz, uint16_t AccHz_mSec)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - StepperConfigSpeedAcc(App) called", __FILENAME__, __LINE__);
  //  call IotModuleSte.ConfigSpeedAcc
  return IotModuleSte.ConfigSpeedAcc(AllPinsData.PinsParams[PinIndex].Channel2, SpeedHz, AccHz_mSec);
}

void cIotModuleApp::PreFuota(void)
{
  //  Stop any interrupt activity
  IotModuleAdc.PreFuota();
  IotModuleSWEncoder.PreFuota();
  IotModulePeriod.PreFuota();
}

void cIotModuleApp::SetModuleName(char *pName)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - name : %s", __FILENAME__, __LINE__, pName);

  nvs_handle my_nvshandle;
  esp_err_t err = nvs_open_from_partition("nvs", "IotModule", NVS_READWRITE, &my_nvshandle);
  //  esp_err_t err = nvs_open("IotModule", NVS_READWRITE, &my_nvshandle);
  if (err == ESP_OK)
  {
    err = nvs_set_str(my_nvshandle, "ModuleName", pName);
    if (err != ESP_OK)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_set_str error", __FILENAME__, __LINE__);
    }
    nvs_close(my_nvshandle);
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_open error", __FILENAME__, __LINE__);
  }
}

void cIotModuleApp::GetModuleName(char *pName, const uint8_t NameSize)
{
  char pDefaultName[] = "NoName";
  nvs_handle my_nvshandle;
  esp_err_t err = nvs_open_from_partition("nvs", "IotModule", NVS_READWRITE, &my_nvshandle);
  //  esp_err_t err = nvs_open("IotModule", NVS_READWRITE, &my_nvshandle);
  if (err == ESP_OK)
  {
    size_t Size = NameSize;
    err = nvs_get_str(my_nvshandle, "ModuleName", pName, &Size);
    if (err != ESP_OK)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_get_str error", __FILENAME__, __LINE__);
      strcpy((char*)pName, pDefaultName);
      err = nvs_set_str(my_nvshandle, "ModuleName", pName);
      if (err != ESP_OK)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_set_str error", __FILENAME__, __LINE__);
      }
    }
    nvs_close(my_nvshandle);
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - nvs_open error", __FILENAME__, __LINE__);
  }
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - name : %s", __FILENAME__, __LINE__, pName);
  if (err != ESP_OK)
  {
    strcpy((char*)pName, pDefaultName);
  }
}

uint8_t IRAM_ATTR cIotModuleApp::GetOutputPinsNumber(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  return AllPinsData.OutputPinsNumber;
}

uint8_t IRAM_ATTR cIotModuleApp::GetInputPinsNumber(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  return AllPinsData.InputPinsNumber;
}

char IRAM_ATTR cIotModuleApp::GetProgStatus(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  char retval = 'N';
  if ((ProgStatus & MODULE) == MODULE)
  {
    retval = 'Y';
  }
  //  if((IOTMODULE_ADC24_PIN_MASTER == 0) && (ProgStatus == MODULE))
  //  {
  //  retval = 'Y';
  //  }
  //  else if((IOTMODULE_ADC24_PIN_MASTER == 1) && (ProgStatus == (MODULE | AD7124)))
  //  {
  //  retval = 'Y';
  //  }
  return retval;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function update the output values on the esp32 hardware
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleApp::SetOutputValues(bool Reset)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  if (GetProgStatus() == 'Y')
  {
    for (uint8_t IndexCount = 0; IndexCount < AllPinsData.OutputPinsNumber; IndexCount++)
    {
      uint8_t Index = AllPinsData.OutputIndexes[IndexCount];
      switch (AllPinsData.PinsParams[Index].Type)
      {
        case DIG_OUT:
          if (Reset) AllPinsData.PinsParams[Index].fValue = 0.0f;
          ESP_ERROR_CHECK(gpio_set_level((gpio_num_t)AllPinsData.PinsParams[Index].Name, AllPinsData.PinsParams[Index].fValue == 0.0f ? 0 : 1));
          break;
        case SERVO:
          //  here the fValue is microseconds from 0 to 4000 
          //  the SERVO pins are fixed to 50 Hz so the PWM period is 20000 uS
          //  fValue == 0 Sleep / default range from 500 uS to 2500 uS
          if (Reset) AllPinsData.PinsParams[Index].fValue = 0.0f;
          IotModuleServo.SetDuty(AllPinsData.PinsParams[Index].Channel2, AllPinsData.PinsParams[Index].fValue);
          break;
        case PWM:
          //  here the fValue is a number from 0.0f to 1.0f representing rapresenting duty from 0% to 100%
          if (Reset) AllPinsData.PinsParams[Index].fValue = 0.0f;
          IotModulePWM.SetDuty(AllPinsData.PinsParams[Index].Channel2, AllPinsData.PinsParams[Index].fValue);
          break;
        case DAC_8:
          if (Reset) AllPinsData.PinsParams[Index].fValue = 0.0f;
          ESP_ERROR_CHECK(dac_output_voltage((dac_channel_t)ModuleProps.PinsProps[Index].Channel1, (uint8_t)(AllPinsData.PinsParams[Index].fValue)));
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - DAC_8 - Index : %d, channel : %d , fValue : %f", __FILENAME__, __LINE__, Index, ModuleProps.PinsProps[Index].Channel1, AllPinsData.PinsParams[Index].fValue);
          break;
          //  generic output values are written by genericWrite* functions
        case GEN_OUT_8:
        case GEN_OUT_16:
        case GEN_OUT_24:
        case GEN_OUT_FLOAT:
          if (Reset) AllPinsData.PinsParams[Index].fValue = 0.0f;
          break;
        case STEPPER:
          if (Reset)
          {
            //  stop movement
            IotModuleSte.Stop(AllPinsData.PinsParams[Index].Channel2);
            //  set reset request  
            AllPinsData.PinsParams[Index].iValue = 0xffffffff;
          }
          else
          {
            IotModuleSte.SetDestination(AllPinsData.PinsParams[Index].Channel2, AllPinsData.PinsParams[Index].iValue);
          }
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Destination : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[Index].iValue);
          break;
        default:
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - output pin type not handled - Index :%d, Type :%d", __FILENAME__, __LINE__, Index, AllPinsData.PinsParams[Index].Type);
          break;
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function update the input values of the esp32 hardware
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleApp::GetInputValues(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  //  if module programmed
  if (GetProgStatus() == 'Y')
  {
    //  if CAP pin configured cycle between them every time this function is requested
    if (AllPinsData.CapNumber > 1)
    {
      AllPinsData.CapSeqCounter++;
      if (AllPinsData.CapSeqCounter == AllPinsData.CapNumber)
      {
        AllPinsData.CapSeqCounter = 0;
      }
    }
    uint8_t AdcChannelIndex = 0;
    uint16_t AdcActiveChannels = IotModuleAdc.GetActiveChannels();
    //  for all the input pins
    for (uint8_t IndexCount = 0; IndexCount < AllPinsData.InputPinsNumber; IndexCount++)
    {
      uint8_t Index = AllPinsData.InputIndexes[IndexCount];
      switch (AllPinsData.PinsParams[Index].Type)
      {
        case DIG_IN:
        case DIG_IN_PU:
          AllPinsData.PinsParams[Index].fValue = gpio_get_level((gpio_num_t)AllPinsData.PinsParams[Index].Name) == 0 ? 0.0 : 1.0;
          break;
        case ADC_16:
          {
            uint32_t tmp = 0;
            for (uint8_t i = 0; i < ADC_CONVERSION_OVERSAMPLING; i++)
            {
              tmp += esp_adc_cal_raw_to_voltage(analogRead(AllPinsData.PinsParams[Index].Name), (const esp_adc_cal_characteristics_t *)&AdcChars);
            }
            AllPinsData.PinsParams[Index].fValue = (float)tmp * adcMult + adcShift;
          }
          break;
        case CAP_16:
          //  ----------------------------------------------------------------------------------------------
          //  Correction 1
          //  - Sometimes there are "glitches" to zero or low values
          //  - Each channel is read three times and the highest value is used (two times is not enought)
          //  ----------------------------------------------------------------------------------------------
          //  Correction 2
          //  - The Pin32 or Pin33 used alone (with no other CAP_16 configured) is always "1"
          //  - When reading as CAP_16 a single Pin32 we read the Pin33 before
          //  - When reading as CAP_16 a single Pin33 we read the Pin32 before
          //  ----------------------------------------------------------------------------------------------
          //  The touchRead function is slow (460 uS), only one pin at a time is read every xchg cycle
          //  It would be better but uselessly complex to prepare all the values in a separate thread
          //  Is better to use a custom function that simply WORKS and is faster than this, if exist
          //  ----------------------------------------------------------------------------------------------
          if (AllPinsData.CapSeqCounter == AllPinsData.PinsParams[Index].CapSeq)
          {
            uint8_t n1;
            uint8_t n2;
            uint8_t name;
            name = ModuleProps.PinsProps[Index].CapName;
            if (AllPinsData.CapNumber == 1)
            {
              if (name == 32) n1 = touchRead(33);
              if (name == 33) n1 = touchRead(32);
            }
            n1 = touchRead(name);
            n2 = touchRead(name);
            if (n2 > n1) n1 = n2;
            n2 = touchRead(name);
            if (n2 > n1) n1 = n2;
            AllPinsData.PinsParams[Index].fValue = n1 * 256.0f;
          }
          break;
        case COUNTER:
        case COUNTER_PU:
          {
            AllPinsData.PinsParams[Index].fValue = IotModuleCounter.GetCount(AllPinsData.PinsParams[Index].Channel2);
          }
          break;
        case PERIOD:
        case PERIOD_PU:
          {
            uint32_t Count;
            IotModulePeriod.GetCount(AllPinsData.PinsParams[Index].Channel2, &Count);
            AllPinsData.PinsParams[Index].fValue = Count;
          }
          break;
        case ENCODER_A:
        case ENCODER_A_PU:
          {
            int16_t Count = 0;
#if IOTMODULE_ENCODERS_TYPE == 0
            AllPinsData.PinsParams[Index].fValue = IotModuleHWEncoder.GetCount(AllPinsData.PinsParams[Index].Channel2);
#elif IOTMODULE_ENCODERS_TYPE == 1
            IotModuleSWEncoder.GetCount(AllPinsData.PinsParams[Index].Channel2, &Count);
#endif
            AllPinsData.PinsParams[Index].fValue = (uint16_t)Count;
          }
          break;
        case ADC_24:
          AllPinsData.PinsParams[Index].fValue = IotModuleAdc.GetISR_Counter();
          break;
        case ADC_24_CH:
          //  if at least there is one channel active
          if (AdcActiveChannels)
          {
            //  while not active channels
            while (!(AdcActiveChannels & 0x1))
            {
              //  skip not active channels  
              AdcActiveChannels >>= 1;
              AdcChannelIndex++;
            }
          }
          //  get integer conversion value for this active channel converting it to float
          AllPinsData.PinsParams[Index].fValue = IotModuleAdc.GetISR_Channel_uint32_Value(AdcChannelIndex++);
          AdcActiveChannels >>= 1;
          break;
          //  generic input values are written by genericRead(*) functions
        case GEN_IN_8:
        case GEN_IN_16:
        case GEN_IN_24:
        case GEN_IN_FLOAT:
          break;
        case STEPPER_DIR:
          AllPinsData.PinsParams[Index].iValue = IotModuleSte.GetDistanceFromDestination(AllPinsData.PinsParams[Index].Channel2);
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Destination : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[Index].iValue);
          break;
          //  not implemented
        default:
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - input pin type not handled - Index :%d, Type :%d", __FILENAME__, __LINE__, Index, AllPinsData.PinsParams[Index].Type);
          break;
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                     RPC functions
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::RPCSetModuleName(const uint8_t ModuleNameSize, uint8_t *pModuleName)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  char pName[32 + 1];
  strncpy((char*)pName, (char*)pModuleName, ModuleNameSize);
  pName[ModuleNameSize] = '\0';
  SetModuleName(pName);
}

void cIotModuleApp::RPCGetModuleName(uint8_t *pModuleNameSize, uint8_t *pModuleName)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);
  char pName[32 + 1];
  GetModuleName(pName, sizeof(pName));
  uint8_t tmp;
  tmp = strlen(pName);
  *pModuleNameSize = tmp;
  strncpy((char*)pModuleName, pName, tmp);
}

uint8_t cIotModuleApp::RPCGetTotalPinsNumber(void)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called, TotalPinsNumber : %d", __FILENAME__, __LINE__, AllPinsData.TotalPinsNumber);
  return AllPinsData.TotalPinsNumber;
}

uint8_t cIotModuleApp::RPCGetPinName(const uint8_t PinIndex)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called, PinIndex : %d, Name : %d", __FILENAME__, __LINE__, PinIndex, ModuleProps.PinsProps[PinIndex].Name);
  return ModuleProps.PinsProps[PinIndex].Name;
}

void cIotModuleApp::RPCGetScanPinTypes(const uint8_t PinIndex, uint8_t *pRPCBuffer)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called, PinIndex : %d", __FILENAME__, __LINE__, PinIndex);
  memcpy(pRPCBuffer, &ModuleProps.PinsProps[PinIndex].Types, ModuleProps.PinsProps[PinIndex].TypesSize);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function get all the possible pin types for a certain pin and removes some types based on the
//  remaining resources of the moment
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::RPCGetDynamicPinTypes(const uint8_t PinIndex, uint8_t *pTotalPinTypes, uint8_t *pPinType)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called - PinIndex : %d", __FILENAME__, __LINE__, PinIndex);

  if (PinIndex < AllPinsData.TotalPinsNumber)
  {
    uint8_t Index = 0;
    bool Found = false;

    if (PinIndex > 0)
    {
      //  if previous pin is encodera then this pin must be only encoderb
      if (AllPinsData.PinsParams[PinIndex - 1].Type == ENCODER_A)
      {
        pPinType[Index++] = ENCODER_B;
        Found = true;
      }
      //  if previous pin is encoderapu then this pin must be only encoderbpu
      else if (AllPinsData.PinsParams[PinIndex - 1].Type == ENCODER_A_PU)
      {
        pPinType[Index++] = ENCODER_B_PU;
        Found = true;
      }
      //  if previous pin is stepper then this pin must be only stepperdir
      else if (AllPinsData.PinsParams[PinIndex - 1].Type == STEPPER)
      {
        pPinType[Index++] = STEPPER_DIR;
        Found = true;
      }
    }
    //  otherwise
    if (!Found)
    {
      //  for all the possible pin types of this pin
      for (uint8_t Index1 = 0; Index1 < ModuleProps.PinsProps[PinIndex].TypesSize; Index1++)
      {
        PIN_TYPE Type = ModuleProps.PinsProps[PinIndex].Types[Index1];
        switch (Type)
        {
          //  add servo or pwm to pin types:
          //  - if ledc channels are available or
          //  - if pin is already configured in this way
          case SERVO:
            if (IotModuleRes.MoreServo() || (AllPinsData.PinsParams[PinIndex].Type == SERVO))
            {
              pPinType[Index++] = Type;
            }
            break;
          //  add servo or pwm to pin types:
          //  - if ledc channels are available or
          //  - if pin is already configured in this way
          case PWM:
            if (IotModuleRes.MorePWM() || (AllPinsData.PinsParams[PinIndex].Type == PWM))
            {
              pPinType[Index++] = Type;
            }
            break;
          //  add counter(pu) to pin types:
          //  - if counters channels are available or
          //  - if pin is already configured in this way
          case COUNTER:
          case COUNTER_PU:
            if (IotModuleRes.MoreCounter() || (AllPinsData.PinsParams[PinIndex].Type == COUNTER) || (AllPinsData.PinsParams[PinIndex].Type == COUNTER_PU))
            {
              pPinType[Index++] = Type;
            }
            break;
          //  add period(pu) to pin types:
          //  - if period channels are available or
          //  - if pin is already configured in this way
          case PERIOD:
          case PERIOD_PU:
            if ((PeriodChannelNumber < cIotModulePeriod::MAX_CHANNELS) || (AllPinsData.PinsParams[PinIndex].Type == PERIOD) || (AllPinsData.PinsParams[PinIndex].Type == PERIOD_PU))
            {
              pPinType[Index++] = Type;
            }
            break;
          //  add encodera(pu) to pin types:
          //  - if counters channels are available and this pin is not encodera(pu) or
          //  - if pin is already configured in this way
          case ENCODER_A:
          case ENCODER_A_PU:
#if IOTMODULE_ENCODERS_TYPE == 0
            if ((IotModuleRes.MoreHWEncoder() && (AllPinsData.PinsParams[PinIndex].Type != ENCODER_A) && (AllPinsData.PinsParams[PinIndex].Type != ENCODER_A_PU)) ||
                ((AllPinsData.PinsParams[PinIndex].Type == ENCODER_A) || (AllPinsData.PinsParams[PinIndex].Type == ENCODER_A_PU)))
#elif IOTMODULE_ENCODERS_TYPE == 1
            if (((SWEncoderChannelNumber < cIotModuleSWEncoder::MAX_CHANNELS) && (AllPinsData.PinsParams[PinIndex].Type != ENCODER_A) && (AllPinsData.PinsParams[PinIndex].Type != ENCODER_A_PU)) ||
                ((AllPinsData.PinsParams[PinIndex].Type == ENCODER_A) || (AllPinsData.PinsParams[PinIndex].Type == ENCODER_A_PU)))
#endif
            {
              pPinType[Index++] = Type;
            }
            break;
          //  add encoderb(pu) to pin types if previous pin is encodera(pu)
          case ENCODER_B:
          case ENCODER_B_PU:
            if ((AllPinsData.PinsParams[PinIndex - 1].Type == ENCODER_A) || (AllPinsData.PinsParams[PinIndex - 1].Type == ENCODER_A_PU))
            {
              pPinType[Index++] = (uint8_t)Type;
            }
            break;
          //  add ADC_24_CH_B to pin types:
          //  - if previous is already programmed as ADC_24_CH and
          //  - the name is even (62, 64, 66...)
          case ADC_24_CH_B:
            if ((AllPinsData.PinsParams[PinIndex - 1].Type == ADC_24_CH) && ((AllPinsData.PinsParams[PinIndex].Name % 2 == 0)))
            {
              pPinType[Index++] = (uint8_t)Type;
            }
            break;
          //  add stepper to pin types:
          //  - if stepper channels are available and this pin is not stepper
          //  - if pin is already configured in this way
          case STEPPER:
            if ((IotModuleRes.MoreStepper() && (AllPinsData.PinsParams[PinIndex].Type != STEPPER)) ||
                 ((AllPinsData.PinsParams[PinIndex].Type == STEPPER)))
            {
              pPinType[Index++] = Type;
            }
            break;
          //  add stepperdir to pin types if previous pin is stepper
          case STEPPER_DIR:
            if (AllPinsData.PinsParams[PinIndex - 1].Type == STEPPER)
            {
              pPinType[Index++] = (uint8_t)Type;
            }
            break;
          //  add always the remaining types
          default:
            pPinType[Index++] = (uint8_t)Type;
            break;
        }
      }
    }
    //  Set number of valid pin types
    *pTotalPinTypes = Index;
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TotalPinTypes : %d", __FILENAME__, __LINE__, Index);
    for (uint8_t i = 0; i < Index; i++)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Type : %d", __FILENAME__, __LINE__, pPinType[i]);
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin number mismatch", __FILENAME__, __LINE__);
    *pTotalPinTypes = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function programs the esp32 hardware to get the pins ready to handle the programmed function
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::RPCSetAllPinsParams(const uint8_t TotalPinsNumber, const uint8_t *pPinType)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  AllPinsData.OutputPinsNumber = AllPinsData.InputPinsNumber = 0;
  AllPinsData.CapNumber = AllPinsData.CapSeqCounter = 0;
  SWEncoderChannelNumber = 0;
  PeriodChannelNumber = 0;
  uint8_t UnusedPinNumber = 0;
  uint8_t NumST = 0, NumSE = 0, NumPW = 0, NumCN = 0, NumEN = 0;

  //  if total within limits
  if (TotalPinsNumber <= AllocatedPinsNumber)
  {
    //  phase 1 : remove the old pin function and restore the hard reset pin status
    //  for all pins
    for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
    {
      //  previous pin type programmed
      switch (AllPinsData.PinsParams[Index].Type)
      {
        //  todoo
        case PWM:
          //  detach
          IotModulePWM.Detach(AllPinsData.PinsParams[Index].Channel2);
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          break;
        case SERVO:
          //  detach
          IotModuleServo.Detach(AllPinsData.PinsParams[Index].Channel2);
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          break;
        case DIG_OUT:
        case DIG_IN:
        case DIG_IN_PU:
        case ADC_16:
        case CAP_16:
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          break;
        case COUNTER:
        case COUNTER_PU:
          //  detach
          IotModuleCounter.Detach(AllPinsData.PinsParams[Index].Channel2);
          break;
          break;
        case DAC_8:
          //  disable the pin DAC functionality
          ESP_ERROR_CHECK(dac_output_disable((dac_channel_t)ModuleProps.PinsProps[Index].Channel1));
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          break;
        case ENCODER_A:
        case ENCODER_A_PU:
#if IOTMODULE_ENCODERS_TYPE == 0
          //  detach A and B
          IotModuleHWEncoder.Detach(AllPinsData.PinsParams[Index].Channel2);
#elif IOTMODULE_ENCODERS_TYPE == 1
          //  reset only A
          IotModuleSWEncoder.Detach(AllPinsData.PinsParams[Index].Channel2);
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
#endif
          break;
        case ENCODER_B:
        case ENCODER_B_PU:
#if IOTMODULE_ENCODERS_TYPE == 0
#elif IOTMODULE_ENCODERS_TYPE == 1
          //  reset B
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
#endif
          break;
        case PERIOD:
        case PERIOD_PU:
          //  detach
          IotModulePeriod.Detach(AllPinsData.PinsParams[Index].Channel2);
          ESP_ERROR_CHECK(gpio_reset_pin((gpio_num_t)AllPinsData.PinsParams[Index].Name));
          break;
        case STEPPER:
          //  detach step and dir
          IotModuleSte.Detach(AllPinsData.PinsParams[Index].Channel2);
          break;
        case ADC_24:
          //  if now not programmed
          if((PIN_TYPE)pPinType[Index] == UNUSED)
          {
            //  detach Adc
            IotModuleAdc.Detach();
          }
          break;
        default:
          //  do nothing for other types
          break;
      }
    }
    //  phase 2 : count new pin type with limited resources
    //  for all pins    
    for (uint8_t Index = 0; Index < TotalPinsNumber; Index++)
    {
      //  get new type
      switch ((PIN_TYPE)pPinType[Index])
      {
        case PWM:
          NumPW++;
          break;
        case SERVO:
          NumSE++;
          break;
        case COUNTER:
        case COUNTER_PU:
          NumCN++;
          break;
        case ENCODER_A:
        case ENCODER_A_PU:
#if IOTMODULE_ENCODERS_TYPE == 0
          NumEN++;
          break;
#elif IOTMODULE_ENCODERS_TYPE == 1
#endif
        case STEPPER:
          NumST++;
          break;
        default:
          //  do nothing for other types
          break;
      }
    }
    //  fill resources
    if(!IotModuleRes.Fill(NumST, NumSE, NumPW, NumCN, NumEN))
    {
      //  todoo error
    }
    //  phase 3 : set the new pin function
    //  for all pins    
    for (uint8_t Index = 0; Index < AllPinsData.TotalPinsNumber; Index++)
    {
      //  get new type
      PIN_TYPE Type = (PIN_TYPE)pPinType[Index];
      //  set the programmed name
      AllPinsData.PinsParams[Index].Name = ModuleProps.PinsProps[Index].Name;
      //  Check if this type is one of the possible types the pin can have
      bool ValidType = false;
      for (uint8_t i = 0; i < ModuleProps.PinsProps[Index].TypesSize; i++)
      {
        if (Type == ModuleProps.PinsProps[Index].Types[i])
        {
          ValidType = true;
          break;
        }
      }
      //  if not valid set it unused
      if (!ValidType)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin type not valid - pin name : %d - type : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[Index].Name, Type);
        Type = UNUSED;
      }
      switch (Type)
      {
        case UNUSED:
          UnusedPinNumber++;
          break;

        case DIG_OUT:
          ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_MODE_OUTPUT));
          ESP_ERROR_CHECK(gpio_set_level((gpio_num_t)AllPinsData.PinsParams[Index].Name, AllPinsData.PinsParams[Index].fValue == 0.0f ? 0 : 1));
          AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
          AllPinsData.OutputPinsNumber++;
          break;

        case SERVO:
          {
            uint8_t LedcChannel = 0, LedcTimer = 0;
            if (IotModuleRes.AllocServo(&LedcChannel, &LedcTimer))
            {
              ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_MODE_OUTPUT));
              //  attach
              IotModuleServo.Attach(LedcChannel, LedcTimer, (gpio_num_t)AllPinsData.PinsParams[Index].Name);
            }
            AllPinsData.PinsParams[Index].Channel2 = LedcChannel;
            AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
            AllPinsData.OutputPinsNumber++;
          }
          break;

        case PWM:
          {
            uint8_t LedcChannel = 0, LedcTimer = 0;
            if (IotModuleRes.AllocPWM(&LedcChannel, &LedcTimer))
            {
              ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_MODE_OUTPUT));
              //  attach
              IotModulePWM.Attach(LedcChannel, LedcTimer, (gpio_num_t)AllPinsData.PinsParams[Index].Name);
            }
            AllPinsData.PinsParams[Index].Channel2 = LedcChannel;
            AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
            AllPinsData.OutputPinsNumber++;
          }
          break;

        case DAC_8:
          ESP_ERROR_CHECK(dac_output_enable((dac_channel_t)ModuleProps.PinsProps[Index].Channel1));
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - DAC_8 - channel : %d", __FILENAME__, __LINE__, ModuleProps.PinsProps[Index].Channel1);
          AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
          AllPinsData.OutputPinsNumber++;
          break;

        case DIG_IN:
        case DIG_IN_PU:
          ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_MODE_INPUT));
          
          if (Type == DIG_IN) 
          {
            ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_FLOATING));
          }
          else
          {
            ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_PULLUP_ONLY));
          }
          
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case ADC_16:
          AllPinsData.PinsParams[Index].Channel2 = ModuleProps.PinsProps[Index].Channel1;
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case CAP_16:
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          AllPinsData.PinsParams[Index].CapSeq = AllPinsData.CapNumber;
          AllPinsData.CapNumber++;
          break;

        case COUNTER:
        case COUNTER_PU:
          {
            uint8_t PcntChannel = 0;
            if (IotModuleRes.AllocCounter(&PcntChannel))
            {
              if (Type == COUNTER)  IotModuleCounter.SetControlPin(PcntChannel, (gpio_num_t)ModuleProps.PinsProps[Index].Name, GPIO_FLOATING, ModuleProps.PinsProps[Index].ResetMode);
              else                  IotModuleCounter.SetControlPin(PcntChannel, (gpio_num_t)ModuleProps.PinsProps[Index].Name, GPIO_PULLUP_ONLY, ModuleProps.PinsProps[Index].ResetMode);
              //  attach
              IotModuleCounter.Attach(PcntChannel);
            }
            AllPinsData.PinsParams[Index].Channel2 = PcntChannel;
            AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
            AllPinsData.InputPinsNumber++;
          }
          break;

        case PERIOD:
        case PERIOD_PU:
          if (PeriodChannelNumber < cIotModulePeriod::MAX_CHANNELS)
          {
            ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_MODE_INPUT));

            if (Type == PERIOD)
            {
              ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_FLOATING));
            }
            else
            {
              ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)AllPinsData.PinsParams[Index].Name, GPIO_PULLUP_ONLY));
            }

            IotModulePeriod.Attach(PeriodChannelNumber, ModuleProps.PinsProps[Index].Name);
            AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
            AllPinsData.PinsParams[Index].Channel2 = PeriodChannelNumber;
            AllPinsData.InputPinsNumber++;
            PeriodChannelNumber++;
          }
          else
          {
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Periods error - PeriodChannelNumber overflow", __FILENAME__, __LINE__);
          }
          break;

        case ENCODER_A:
        case ENCODER_A_PU:
#if IOTMODULE_ENCODERS_TYPE == 0
          {
            uint8_t PcntChannel = 0;
            if (IotModuleRes.AllocHWEncoder(&PcntChannel))
            {
              //  set step GPIO
              switch (Type)
              {
                case ENCODER_A:
                  //  set GPIO
                  IotModuleHWEncoder.SetControlPin(PcntChannel, cIotModuleHWEncoder::PIN_A, (gpio_num_t)ModuleProps.PinsProps[Index].Name, GPIO_FLOATING, ModuleProps.PinsProps[Index].ResetMode);
                  //  set GPIO
                  IotModuleHWEncoder.SetControlPin(PcntChannel, cIotModuleHWEncoder::PIN_B, (gpio_num_t)ModuleProps.PinsProps[Index + 1].Name, GPIO_FLOATING, ModuleProps.PinsProps[Index].ResetMode);
                  break;
                case ENCODER_A_PU:
                  //  set GPIO
                  IotModuleHWEncoder.SetControlPin(PcntChannel, cIotModuleHWEncoder::PIN_A, (gpio_num_t)ModuleProps.PinsProps[Index].Name, GPIO_PULLUP_ONLY, ModuleProps.PinsProps[Index].ResetMode);
                  //  set GPIO
                  IotModuleHWEncoder.SetControlPin(PcntChannel, cIotModuleHWEncoder::PIN_B, (gpio_num_t)ModuleProps.PinsProps[Index + 1].Name, GPIO_PULLUP_ONLY, ModuleProps.PinsProps[Index].ResetMode);
                  break;
                default:
                  break;
              }
              //  attach
              IotModuleHWEncoder.Attach(PcntChannel);
            }
            AllPinsData.PinsParams[Index].Channel2 = PcntChannel;
            AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
            AllPinsData.InputPinsNumber++;
          }
#elif IOTMODULE_ENCODERS_TYPE == 1
          if (SWEncoderChannelNumber < cIotModuleSWEncoder::MAX_CHANNELS)
          {
            //  set the programmed names
            AllPinsData.PinsParams[Index].Name = ModuleProps.PinsProps[Index].Name;
            AllPinsData.PinsParams[Index + 1].Name = ModuleProps.PinsProps[Index + 1].Name;
            ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)ModuleProps.PinsProps[Index].Name, GPIO_MODE_INPUT));
            ESP_ERROR_CHECK(gpio_set_direction((gpio_num_t)ModuleProps.PinsProps[Index + 1].Name, GPIO_MODE_INPUT));
            switch (Type)
            {
              case ENCODER_A:
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)ModuleProps.PinsProps[Index].Name, GPIO_FLOATING));
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)ModuleProps.PinsProps[Index + 1].Name, GPIO_FLOATING));
                //  set the next pin type
                AllPinsData.PinsParams[Index + 1].Type = ENCODER_B;
                break;
              case ENCODER_A_PU:
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)ModuleProps.PinsProps[Index].Name, GPIO_PULLUP_ONLY));
                ESP_ERROR_CHECK(gpio_set_pull_mode((gpio_num_t)ModuleProps.PinsProps[Index + 1].Name, GPIO_PULLUP_ONLY));
                //  set the next pin type
                AllPinsData.PinsParams[Index + 1].Type = ENCODER_B_PU;
                break;
              default:
                break;
            }
            IotModuleSWEncoder.Attach(SWEncoderChannelNumber, (gpio_num_t)ModuleProps.PinsProps[Index].Name, (gpio_num_t)ModuleProps.PinsProps[Index + 1].Name);
            AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
            AllPinsData.PinsParams[Index].Channel2 = SWEncoderChannelNumber;
            AllPinsData.InputPinsNumber++;
            SWEncoderChannelNumber++;
          }
#endif
          break;

        case ENCODER_B:
        case ENCODER_B_PU:
          //  encoder_b dummy pin
          UnusedPinNumber++;
          break;

        case GEN_OUT_8:
        case GEN_OUT_16:
        case GEN_OUT_24:
        case GEN_OUT_FLOAT:
          AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
          AllPinsData.OutputPinsNumber++;
          break;

        case GEN_IN_8:
        case GEN_IN_16:
        case GEN_IN_24:
        case GEN_IN_FLOAT:
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case ADC_24:
          //  if was not programmed
          if(AllPinsData.PinsParams[Index].Type == UNUSED)
          {
            IotModuleAdc.Attach();
          }
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case ADC_24_CH:
          //  adc24 pin used for conversion value
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;

        case ADC_24_CH_B:
          //  adc24 dummy pin
          UnusedPinNumber++;
          break;

        case STEPPER:
          {
            uint8_t LedcTimer = 0, LedcChannel = 0, PcntChannel = 0;
            if (IotModuleRes.AllocStepper(&LedcTimer, &LedcChannel, &PcntChannel))
            {
              //  set step GPIO
              IotModuleSte.SetControlPin(LedcTimer, cIotModuleSte::PIN_STEP, (gpio_num_t)ModuleProps.PinsProps[Index].Name, ModuleProps.PinsProps[Index].ResetMode);
              //  set dir GPIO
              IotModuleSte.SetControlPin(LedcTimer, cIotModuleSte::PIN_DIR, (gpio_num_t)ModuleProps.PinsProps[Index + 1].Name, ModuleProps.PinsProps[Index + 1].ResetMode);
              //  attach
              IotModuleSte.Attach(LedcTimer, LedcChannel, PcntChannel);
              //  set position reset request for attach
              AllPinsData.PinsParams[Index].iValue = 0xffffffff;
            }
            AllPinsData.PinsParams[Index].Channel2 = LedcTimer;
            AllPinsData.PinsParams[Index + 1].Channel2 = LedcTimer;
            AllPinsData.OutputIndexes[AllPinsData.OutputPinsNumber] = Index;
            AllPinsData.OutputPinsNumber++;
          }
          break;

        case STEPPER_DIR:
          AllPinsData.InputIndexes[AllPinsData.InputPinsNumber] = Index;
          AllPinsData.InputPinsNumber++;
          break;
          
      }
      //  set the type
      AllPinsData.PinsParams[Index].Type = Type;
    }

    AllPinsData.TotalPinsNumber = TotalPinsNumber;

    if ((AllPinsData.OutputPinsNumber + AllPinsData.InputPinsNumber + UnusedPinNumber) == TotalPinsNumber)
    {
      //	all module pins programmed
      ProgStatus =  static_cast<PROG_STATUS>(ProgStatus | MODULE);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - all module pins programmed - Tot : %d, In : %d, Out : %d, Unused : %d", __FILENAME__, __LINE__, TotalPinsNumber, AllPinsData.InputPinsNumber, AllPinsData.OutputPinsNumber, UnusedPinNumber);
      for (uint8_t i = 0; i < TotalPinsNumber; i++)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Name : %d, Type : %d", __FILENAME__, __LINE__, AllPinsData.PinsParams[i].Name, AllPinsData.PinsParams[i].Type);
      }
    }
    else
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin mismatch", __FILENAME__, __LINE__);
      //  some pin programmed and some not, pin program Mismatch
      ProgStatus = NOTHING;
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin allocation error", __FILENAME__, __LINE__);
    //  no pin programmed
    AllPinsData.TotalPinsNumber = 0;
    ProgStatus = NOTHING;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function set PWM channel frequency
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::RPCPwmSetFrequency(const uint8_t PinIndex, const float Frequency)
{
  IotModulePWM.SetFreq(AllPinsData.PinsParams[PinIndex].Channel2, Frequency);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function copy the output values from the communication buffer
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleApp::RPCSetOutputXchgValues(const uint8_t IndexCount, const float *pOutValue)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  if (IndexCount < AllPinsData.OutputPinsNumber)
  {
    uint8_t Index = AllPinsData.OutputIndexes[IndexCount];
    switch (AllPinsData.PinsParams[Index].Type)
    {
      case SERVO:
      case PWM:
      case DIG_OUT:
      case DAC_8:
      case GEN_OUT_FLOAT:
      case GEN_OUT_8:
      case GEN_OUT_16:
      case GEN_OUT_24:
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - %d", __FILENAME__, __LINE__, Index);
        AllPinsData.PinsParams[Index].fValue = *pOutValue;
        break;
      case STEPPER:
        {
          uint32_t *pInt32 = (uint32_t *)pOutValue;
          AllPinsData.PinsParams[Index].iValue = *pInt32;
        }
        break;
      default:
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - output pin type not handled - Index :%d, Type :%d", __FILENAME__, __LINE__, Index, AllPinsData.PinsParams[Index].Type);
        //  zero value
        AllPinsData.PinsParams[Index].fValue = 0.0f;
        //  pin program mismatch
        ProgStatus = NOTHING;
        break;
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin number mismatch", __FILENAME__, __LINE__);
    ProgStatus = NOTHING;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//  this function copy the input values to the communication buffer
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleApp::RPCGetInputXchgValues(const uint8_t IndexCount, float *pInValue)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - called", __FILENAME__, __LINE__);

  if (IndexCount < AllPinsData.InputPinsNumber)
  {
    uint8_t Index = AllPinsData.InputIndexes[IndexCount];
    switch (AllPinsData.PinsParams[Index].Type)
    {
      case DIG_IN:
      case DIG_IN_PU:
      case ADC_16:
      case CAP_16:
      case COUNTER:
      case COUNTER_PU:
      case PERIOD:
      case PERIOD_PU:
      case ENCODER_A:
      case ENCODER_A_PU:
      case ENCODER_B:
      case ENCODER_B_PU:
      case GEN_IN_8:
      case GEN_IN_16:
      case GEN_IN_24:
      case GEN_IN_FLOAT:
      case ADC_24:
      case ADC_24_CH:
        *pInValue = AllPinsData.PinsParams[Index].fValue;
        break;
      case STEPPER_DIR:
        {
          uint32_t *pInt32 = (uint32_t *)pInValue;
          *pInt32 = AllPinsData.PinsParams[Index].iValue;
        }
        break;
      default:
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - input pin type not handled - Index :%d, Type :%d", __FILENAME__, __LINE__, Index, AllPinsData.PinsParams[Index].Type);
        //  zero value
        *pInValue = 0.0f;
        //  pin program Mismatch
        ProgStatus = NOTHING;
        break;
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - pin number mismatch", __FILENAME__, __LINE__);
    ProgStatus = NOTHING;
  }
}

void cIotModuleApp::TimedInterpolate(void)
{
  IotModuleSte.TimedInterpolate();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                 User program interface functions
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleApp::GenericWrite8 (const uint8_t PinName, const uint8_t Value)
{
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      AllPinsData.PinsParams[Index].fValue = Value;
    }
  }
}

void cIotModuleApp::GenericWrite16(const uint8_t PinName, const uint16_t Value)
{
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      AllPinsData.PinsParams[Index].fValue = Value;
    }
  }
}

void cIotModuleApp::GenericWrite24(const uint8_t PinName, const uint32_t Value)
{
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      AllPinsData.PinsParams[Index].fValue = Value;
    }
  }
}

void IRAM_ATTR cIotModuleApp::GenericWriteFloat(const uint8_t PinName, const float Value)
{
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      AllPinsData.PinsParams[Index].fValue = Value;
    }
  }
}

uint8_t cIotModuleApp::GenericRead8 (const uint8_t PinName)
{
  uint8_t RetValue = 0;
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      RetValue = AllPinsData.PinsParams[Index].fValue;
    }
  }
  return RetValue;
}

uint16_t cIotModuleApp::GenericRead16(const uint8_t PinName)
{
  uint16_t RetValue = 0;
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      RetValue = AllPinsData.PinsParams[Index].fValue;
    }
  }
  return RetValue;
}

uint32_t cIotModuleApp::GenericRead24(const uint8_t PinName)
{
  uint32_t RetValue = 0;
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      RetValue = AllPinsData.PinsParams[Index].fValue;
    }
  }
  return RetValue;
}

float IRAM_ATTR cIotModuleApp::GenericReadFloat(const uint8_t PinName)
{
  float RetValue = 0.0f;
  if (PinName < IOTMODULE_LOOKUP_TABLE_SIZE)
  {
    uint8_t Index = PinIndexLookupTable[PinName];
    if (Index != 255)
    {
      RetValue = AllPinsData.PinsParams[Index].fValue;
    }
  }
  return RetValue;
}
