//================================================================================================
//                                  cIotModuleCom body
//================================================================================================

void cIotModuleCom::Initialize(pWifiEventHandler WifiEventHandler)
{
  const char pSSID[] = IOTMODULE_WIFI_SSID;
  const char pPassword[] = IOTMODULE_WIFI_PASSWORD;
  const char pIPInfo[] = IOTMODULE_STATIC_IP ;
  WIFI_MODE WiFiMode = IOTMODULE_WIFI_MODE == 0 ? STATION : SOFTAP;

  WifiEventHandlerCallback = WifiEventHandler;

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);

  //  Chip Silicon package
  switch ((REG_READ(EFUSE_BLK0_RDATA3_REG) >>  EFUSE_RD_CHIP_VER_PKG_S) & EFUSE_RD_CHIP_VER_PKG_V)
  {
    case EFUSE_RD_CHIP_VER_PKG_ESP32D0WDQ6:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : D0WDQ6", __FILENAME__, __LINE__);
      break;
    case EFUSE_RD_CHIP_VER_PKG_ESP32D0WDQ5:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : D0WDQ5", __FILENAME__, __LINE__);
      break;
    case EFUSE_RD_CHIP_VER_PKG_ESP32D2WDQ5:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : D2WDQ5", __FILENAME__, __LINE__);
      break;
    case EFUSE_RD_CHIP_VER_PKG_ESP32PICOD2:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : PICOD2", __FILENAME__, __LINE__);
      break;
    case EFUSE_RD_CHIP_VER_PKG_ESP32PICOD4:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : PICOD4", __FILENAME__, __LINE__);
      break;
  }

  //  Chip Silicon Revision
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip Silicon Revision : %d", __FILENAME__, __LINE__, (REG_READ(EFUSE_BLK0_RDATA3_REG) >>  EFUSE_RD_CHIP_VER_REV1_S) & EFUSE_RD_CHIP_VER_REV1_V);

  // typedef struct {
  //    rtc_cpu_freq_src_t source;      //!< The clock from which CPU clock is derived
  //    uint32_t source_freq_mhz;       //!< Source clock frequency
  //    uint32_t div;                   //!< Divider, freq_mhz = source_freq_mhz / div
  //    uint32_t freq_mhz;              //!< CPU clock frequency
  // } rtc_cpu_freq_config_t;
  rtc_cpu_freq_config_t rtc_cpu_freq_config;
  rtc_clk_cpu_freq_get_config(&rtc_cpu_freq_config);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - CPU cores frequency : %d Mhz", __FILENAME__, __LINE__, rtc_cpu_freq_config.freq_mhz);
  for (uint8_t i = 0; i < 2; i++)
  {
    /*
    POWERON_RESET           1   Vbat power on reset
    SW_RESET 	              3   Software reset digital core
    OWDT_RESET              4   Legacy watch dog reset digital core
    DEEPSLEEP_RESET         3   Deep Sleep reset digital core
    SDIO_RESET 	            6   Reset by SLC module, reset digital core
    TG0WDT_SYS_RESET        7   Timer Group0 Watch dog reset digital core
    TG1WDT_SYS_RESET        8   Timer Group1 Watch dog reset digital core
    RTCWDT_SYS_RESET        9   RTC Watch dog Reset digital core
    INTRUSION_RESET         10  Instrusion tested to reset CPU
    TGWDT_CPU_RESET         11  Time Group reset CPU
    SW_CPU_RESET 	          12  Software reset CPU
    RTCWDT_CPU_RESET        13  RTC Watch dog Reset CPU
    EXT_CPU_RESET           14  for APP CPU, reseted by PRO CPU
    RTCWDT_BROWN_OUT_RESET 	15  Reset when the vdd voltage is not stable
    RTCWDT_RTC_RESET 	      16  RTC Watch dog reset digital core and rtc module 

    -PRO-  -APP- ------Source------- ----ResetType---

    0x01   0x01  Chip Power On Reset System Reset
    0x10   0x10  RWDT System Reset   System Reset
    0x0F   0x0F  Brown Out Reset     System Reset
    0x03   0x03  Software System     Reset Core Reset
    0x05   0x05  Deep Sleep Reset    Core Reset
    0x07   0x07  MWDT0 Global Reset  Core Reset
    0x08   0x08  MWDT1 Global Reset  Core Reset
    0x09   0x09  RWDT Core Reset     Core Reset
    0x0B         MWDT0 CPU Reset     CPU Reset
    0x0C         Software CPU Reset  CPU Reset
           0x0B  MWDT1 CPU Reset     CPU Reset
           0x0C  Software CPU Reset  CPU Reset
    0x0D   0x0D  RWDT CPU Reset      CPU Reset 
           0x0E  PRO CPU Reset       CPU Reset
    */

    const char *pReason[] =
    {
      "NO_MEAN",
      "POWERON_RESET",
      "NO_MEAN",
      "SW_RESET",
      "OWDT_RESET",
      "DEEPSLEEP_RESET",
      "SDIO_RESET",
      "TG0WDT_SYS_RESET",
      "TG1WDT_SYS_RESET",
      "RTCWDT_SYS_RESET",
      "INTRUSION_RESET",
      "TGWDT_CPU_RESET",
      "SW_CPU_RESET",
      "RTCWDT_CPU_RESET",
      "EXT_CPU_RESET",
      "RTCWDT_BROWN_OUT_RESET",
      "RTCWDT_RTC_RESET",
    };
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Reset reason core %d : %s", __FILENAME__, __LINE__, i, pReason[rtc_get_reset_reason(i)]);
  }

  //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Wakeup reason : %s", __FILENAME__, __LINE__, pReason[rtc_get_wakeup_cause()]);

  //  Cache enabled
  if (spi_flash_cache_enabled())
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - both CPUs have flash cache enabled", __FILENAME__, __LINE__);
  }

  size_t ul;
  ul = spi_flash_get_chip_size();
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Flash chip size : %d bytes", __FILENAME__, __LINE__, ul);

  //=====================================================================================
  //  flash mapped regions
  //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Flash mapped regions", __FILENAME__, __LINE__);
  //  this function use always serial monitor to dump info, is not ESP_LOG frendly
  //  spi_flash_mmap_dump();
  //=====================================================================================
  //  app_main() -> initArduino() already re-initialize NVS in case of ESP_ERR_NVS_NO_FREE_PAGES
  //  calling esp_partition_find_first(NVS) and esp_partition_erase_range()
  //=====================================================================================
  //  initialize NVS.
  //  esp_err_t err = nvs_flash_init();
  //  if (err == ESP_ERR_NVS_NO_FREE_PAGES /*|| err == ESP_ERR_NVS_NEW_VERSION_FOUND*/)
  //  {
  //    //  OTA app partition table has a smaller NVS partition size than the non-OTA
  //    //  partition table. This size mismatch may cause NVS initialization to fail.
  //    //  If this happens, we erase NVS partition and initialize NVS again.
  //    //  We erase it also if new NVS version is available
  //    ESP_ERROR_CHECK(nvs_flash_erase());
  //    err = nvs_flash_init();
  //  }
  //  //  after init or erase & init there must be NO error left
  //  ESP_ERROR_CHECK(err);
  //=====================================================================================

  //  partitions
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Partition table:", __FILENAME__, __LINE__);
  esp_partition_iterator_t _mypartiterator;
  const esp_partition_t *_mypart;
  _mypartiterator = esp_partition_find(ESP_PARTITION_TYPE_APP, ESP_PARTITION_SUBTYPE_ANY, NULL);
  if (_mypartiterator)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - App Partition table:", __FILENAME__, __LINE__);
    do
    {
      _mypart = esp_partition_get(_mypartiterator);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Type: %02x SubType %02x Address 0x%06X Size 0x%06X Encryption %i Label %s", __FILENAME__, __LINE__, _mypart->type, _mypart->subtype, _mypart->address, _mypart->size, _mypart->encrypted, _mypart->label);
    }
    while ((_mypartiterator = (esp_partition_next(_mypartiterator))));
  }
  esp_partition_iterator_release(_mypartiterator);
  _mypartiterator = esp_partition_find(ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_ANY, NULL);
  if (_mypartiterator)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Data Partition table:", __FILENAME__, __LINE__);
    do
    {
      _mypart = esp_partition_get(_mypartiterator);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Type: %02x SubType %02x Address 0x%06X Size 0x%06X Encryption %i Label %s", __FILENAME__, __LINE__, _mypart->type, _mypart->subtype, _mypart->address, _mypart->size, _mypart->encrypted, _mypart->label);
    }
    while ((_mypartiterator = esp_partition_next(_mypartiterator))) ;
  }
  esp_partition_iterator_release(_mypartiterator);

  //  Running partition
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Running partition : %s", __FILENAME__, __LINE__, esp_ota_get_running_partition()->label);

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Arduino IDE ESP-IDF : %s", __FILENAME__, __LINE__, esp_get_idf_version());
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ESP-IDF portTICK_PERIOD_MS : %d", __FILENAME__, __LINE__, portTICK_PERIOD_MS);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - internal IOTMODULE_FIRMWARE_VER : %d", __FILENAME__, __LINE__, IOTMODULE_FIRMWARE_VER);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - internal IOTMODULE_THEREMINO_IOT_PROTOCOL : %s", __FILENAME__, __LINE__, IOTMODULE_THEREMINO_IOT_PROTOCOL);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - internal IOTMODULE_THEREMINO_IOT_PROTOCOL_VER : %d", __FILENAME__, __LINE__, IOTMODULE_THEREMINO_IOT_PROTOCOL_VER);
  if (IOTMODULE_OUTPUT_WATCHDOG_ENABLE == 1)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_OUTPUT_WATCHDOG_ENABLE : %d", __FILENAME__, __LINE__, IOTMODULE_OUTPUT_WATCHDOG_ENABLE);
  }
  if (IOTMODULE_FUOTA_ENABLE == 1)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_FUOTA_ENABLE : %d", __FILENAME__, __LINE__, IOTMODULE_FUOTA_ENABLE);
  }
  hEventGroup = xEventGroupCreate();
  //  get current execution core ID
  int32_t CoreID = xPortGetCoreID();
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Arduino setup() and loop() execution core : %d", __FILENAME__, __LINE__, CoreID);
  //  if nedeed create ComServerTask pinned to different core
  CoreID = (CoreID == 0 ? 0 : 1);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ComServerTask execution core : %d", __FILENAME__, __LINE__, xPortGetCoreID());
  if (xTaskCreatePinnedToCore(ComServerTask, "udp_server", COM_TASK_STACK_SIZE, this, tskIDLE_PRIORITY + 6, &hComServerTask, CoreID) == pdPASS)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ComServerTask created", __FILENAME__, __LINE__);
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ComServerTask not created", __FILENAME__, __LINE__);
  }

  //  IotModuleApp initialization must be done before any wifi loop event
  IotModuleApp.Initialize();

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_WIFI_SSID : %s", __FILENAME__, __LINE__, pSSID);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_WIFI_PASSWORD : %s", __FILENAME__, __LINE__, pPassword);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_STATIC_IP  : %s", __FILENAME__, __LINE__, pIPInfo);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_WIFI_MODE : %s", __FILENAME__, __LINE__, WiFiMode == STATION ? "Station" : "SoftAP");
  
  //  if configured WiFi mode must be overriden by a pin status
  if(IotModuleApp.WiFiModeOverridePinDefined())
  {
    //  override WiFi mode
    WiFiMode = IotModuleApp.WiFiModeOverridePinStatus() == false ? STATION : SOFTAP;
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - overriden by IOTMODULE_WIFI_MODE_PIN : %s", __FILENAME__, __LINE__, WiFiMode == STATION ? "Station" : "SoftAP");
  }

  //  tcpip_adapter_init required also for UDP
  tcpip_adapter_init();

  //  if static IP
  if (strlen(pIPInfo) != 0)
  {
    char delimiter = ' ';
    std::string tobesplit(pIPInfo);
    std::vector<std::string> tokens;
    std::string token;
    std::istringstream tokenStream(tobesplit);
    while (std::getline(tokenStream, token, delimiter))
    {
      tokens.push_back(token);
    }
    if (tokens.size() == 4)
    {
      tcpip_adapter_ip_info_t AdapterIPInfo;
      tcpip_adapter_dns_info_t DNS_Info;
      ip4addr_aton(tokens[0].c_str(), &AdapterIPInfo.ip);
      ip4addr_aton(tokens[1].c_str(), &AdapterIPInfo.gw);
      ip4addr_aton(tokens[2].c_str(), &AdapterIPInfo.netmask);
      ipaddr_aton(tokens[3].c_str(), &DNS_Info.ip);
      if (WiFiMode == STATION)
      {
        //  stop DHCP client
        ESP_ERROR_CHECK(tcpip_adapter_dhcpc_stop(TCPIP_ADAPTER_IF_STA));
        //  set static IP
        ESP_ERROR_CHECK(tcpip_adapter_set_ip_info(TCPIP_ADAPTER_IF_STA, &AdapterIPInfo));
        //  set DNS IP
        ESP_ERROR_CHECK(tcpip_adapter_set_dns_info(TCPIP_ADAPTER_IF_STA, TCPIP_ADAPTER_DNS_MAIN, &DNS_Info));
      }
      else if (WiFiMode == SOFTAP)
      {
        //  stop DHCP server
        ESP_ERROR_CHECK(tcpip_adapter_dhcps_stop(TCPIP_ADAPTER_IF_AP));
        //  set static IP
        ESP_ERROR_CHECK(tcpip_adapter_set_ip_info(TCPIP_ADAPTER_IF_AP, &AdapterIPInfo));
        //  set DNS IP
        ESP_ERROR_CHECK(tcpip_adapter_set_dns_info(TCPIP_ADAPTER_IF_AP, TCPIP_ADAPTER_DNS_MAIN, &DNS_Info));
        //  start DHCP server
        ESP_ERROR_CHECK(tcpip_adapter_dhcps_start(TCPIP_ADAPTER_IF_AP));
      }
    }
    else
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Static IP parameters not valid, network mode forced to DHCP", __FILENAME__, __LINE__);
    }
  }

  ESP_ERROR_CHECK(esp_event_loop_init(cIotModuleCom::WifiEventHandler, this));
  wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
  ESP_ERROR_CHECK(esp_wifi_init(&cfg));
  ESP_ERROR_CHECK(esp_wifi_set_storage(WIFI_STORAGE_RAM));
  //  power save must be disabled to achive fastest WiFi performace
  ESP_ERROR_CHECK(esp_wifi_set_ps(WIFI_PS_NONE));

  if (WiFiMode == STATION)
  {
    strcpy(reinterpret_cast<char*>(WifiConfig.sta.ssid), pSSID);
    strcpy((char*)WifiConfig.sta.password, pPassword);
    WifiConfig.sta.bssid_set = 0;
    WifiConfig.sta.channel = 0;
    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_STA ));
    ESP_ERROR_CHECK(esp_wifi_set_config(ESP_IF_WIFI_STA, &WifiConfig));
  }
  else if (WiFiMode == SOFTAP)
  {
    strcpy(reinterpret_cast<char*>(WifiConfig.ap.ssid), pSSID);
    WifiConfig.ap.ssid_len = 0;
    strcpy((char*)WifiConfig.ap.password, pPassword);
    WifiConfig.ap.authmode = WIFI_AUTH_WPA2_PSK;
    WifiConfig.ap.max_connection = 1;
    WifiConfig.ap.channel = IOTMODULE_WIFI_CHANNEL;
    WifiConfig.ap.ssid_hidden = 0;
    WifiConfig.ap.beacon_interval = 100;
    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_AP ));
    ESP_ERROR_CHECK(esp_wifi_set_config(ESP_IF_WIFI_AP, &WifiConfig));
  }
  ESP_ERROR_CHECK(esp_wifi_start());

  if (WiFiMode == STATION)
  {
    ESP_ERROR_CHECK(esp_wifi_get_mac(WIFI_IF_STA, WiFiMAC));
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Station mode MAC address : %02X%02X%02X%02X%02X%02X", __FILENAME__, __LINE__, WiFiMAC[0], WiFiMAC[1], WiFiMAC[2], WiFiMAC[3], WiFiMAC[4], WiFiMAC[5]);
  }
  else if (WiFiMode == SOFTAP)
  {
    ESP_ERROR_CHECK(esp_wifi_get_mac(WIFI_IF_AP, WiFiMAC));
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - AP mode MAC address : %02X%02X%02X%02X%02X%02X", __FILENAME__, __LINE__, WiFiMAC[0], WiFiMAC[1], WiFiMAC[2], WiFiMAC[3], WiFiMAC[4], WiFiMAC[5]);
  }

  IO_WatchDogCounter = 0;
  FUOTA_WatchDogEnable = false;
  FUOTA_WatchDogCounter = 0;

  //  start ticker activity every 10 mSec
  hTimerCallback = xTimerCreate("Timer", TIMER_MS / portTICK_PERIOD_MS, pdTRUE, this, TimerCallback);
  xTimerStart(hTimerCallback, 0);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                   WiFi events
///////////////////////////////////////////////////////////////////////////////////////////////////////

esp_err_t cIotModuleCom::WifiEventHandler(void *ctx, system_event_t *event)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - WifiEventHandler execution core : %d", __FILENAME__, __LINE__, xPortGetCoreID());

  cIotModuleCom *pIotModuleCom = (cIotModuleCom *)ctx;

  esp_err_t ret = ESP_OK;
  tcpip_adapter_ip_info_t AdapterIPInfo;
  tcpip_adapter_dns_info_t DnsInfo;
  char ipString[15 + 1];

  switch (event->event_id)
  {
    case SYSTEM_EVENT_WIFI_READY:
      //  never seen
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_WIFI_READY", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_SCAN_DONE:
      //  it must never happens
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_SCAN_DONE", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_STA_START:
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_START", __FILENAME__, __LINE__);
        //  get IotModule name
        char pModuleName[32 + 1];
        pIotModuleCom->IotModuleApp.GetModuleName(pModuleName, sizeof(pModuleName));
        //  set IotModule host name
        ESP_ERROR_CHECK(tcpip_adapter_set_hostname(TCPIP_ADAPTER_IF_STA, pModuleName));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Host name : %s", __FILENAME__, __LINE__, pModuleName);
        //  create the UDP socket and begin receive loop in ComServerTask
        if (pIotModuleCom->hComServerTask != nullptr)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_START - xEventGroupSetBits", __FILENAME__, __LINE__);
          xEventGroupSetBits(pIotModuleCom->hEventGroup, 1);
        }
        else
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - EVENT_STA_START - ComServerTask not available", __FILENAME__, __LINE__);
        }
        //  begin to connect to AP
        ESP_ERROR_CHECK(esp_wifi_connect());
      }
      break;

    case SYSTEM_EVENT_STA_STOP:
      //  happens only when FUOTA active and restarting with esp_restart()
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_STOP", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_STA_CONNECTED:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - connected to SSID : %s", __FILENAME__, __LINE__, pIotModuleCom->WifiConfig.ap.ssid);
      wifi_country_t AP_CountryInfo;
      ESP_ERROR_CHECK(esp_wifi_get_country(&AP_CountryInfo));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP CountryInfo - cc : %s", __FILENAME__, __LINE__, AP_CountryInfo.cc);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP CountryInfo - schan : %d", __FILENAME__, __LINE__, AP_CountryInfo.schan);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP CountryInfo - nchan : %d", __FILENAME__, __LINE__, AP_CountryInfo.nchan);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP CountryInfo - max_tx_power : %d", __FILENAME__, __LINE__, AP_CountryInfo.max_tx_power);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP CountryInfo - policy : %s", __FILENAME__, __LINE__, AP_CountryInfo.policy == WIFI_COUNTRY_POLICY_AUTO ? "AUTO" : "MANUAL");
      wifi_ap_record_t AP_Info;
      ESP_ERROR_CHECK(esp_wifi_sta_get_ap_info(&AP_Info));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP Info - channel : %d", __FILENAME__, __LINE__, AP_Info.primary);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP Info - RSSI : %d dBm", __FILENAME__, __LINE__, AP_Info.rssi);
      //  maximum WiFi transmitting power, unit is 0.25dBm, range is [40, 82] corresponding to 10dBm - 20.5dBm
      int8_t MaxTX_Power;
      ESP_ERROR_CHECK(esp_wifi_get_max_tx_power(&MaxTX_Power));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - IotModule MaxTX_Power : %2.2f dBm", __FILENAME__, __LINE__, MaxTX_Power * 0.25f);
      //  code to set iotmodule TX power
      //  MaxTX_Power = 10.0f * 4;
      //  ESP_ERROR_CHECK(esp_wifi_set_max_tx_power(MaxTX_Power));
      //  ESP_ERROR_CHECK(esp_wifi_get_max_tx_power(&MaxTX_Power));
      //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - IotModule MaxTX_Power : %f dBm", __FILENAME__, __LINE__, MaxTX_Power * 0.25f);
      break;

    case SYSTEM_EVENT_STA_DISCONNECTED:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_DISCONNECTED", __FILENAME__, __LINE__);
      //  do not connect when FUOTA active and restarting with esp_restart()
      if (!pIotModuleCom->FUOTA_WatchDogEnable)
      {
        ESP_ERROR_CHECK(esp_wifi_connect());
      }
      break;

    case SYSTEM_EVENT_STA_AUTHMODE_CHANGE:
      //  never tested
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_AUTHMODE_CHANGE", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_STA_GOT_IP:
      //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - IPv4 : " IPSTR " - changed : %d", __FILENAME__, __LINE__, IP2STR(&event->event_info.got_ip.ip_info.ip), event->event_info.got_ip.ip_changed);
      ESP_ERROR_CHECK(tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_STA, &AdapterIPInfo));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_GOT_IP - IP : " IPSTR " , GW : " IPSTR " , NM : " IPSTR ", IP changed : %d", __FILENAME__, __LINE__, IP2STR(&AdapterIPInfo.ip), IP2STR(&AdapterIPInfo.gw),
                          IP2STR(&AdapterIPInfo.netmask), event->event_info.got_ip.ip_changed);
      ESP_ERROR_CHECK(tcpip_adapter_get_dns_info(TCPIP_ADAPTER_IF_STA, TCPIP_ADAPTER_DNS_MAIN, &DnsInfo));
      inet_ntop(AF_INET, &DnsInfo.ip, ipString, sizeof(ipString));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_GOT_IP - DNS - IP : %s", __FILENAME__, __LINE__, ipString);
      break;

    case SYSTEM_EVENT_STA_LOST_IP:
      //  never seen
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_LOST_IP", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_AP_START:
      {
        //  get IotModule name
        char pModuleName[32 + 1];
        pIotModuleCom->IotModuleApp.GetModuleName(pModuleName, sizeof(pModuleName));
        //  set IotModule host name
        ESP_ERROR_CHECK(tcpip_adapter_set_hostname(TCPIP_ADAPTER_IF_AP, pModuleName));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Host name : %s", __FILENAME__, __LINE__, pModuleName);
        
        ESP_ERROR_CHECK(tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_AP, &AdapterIPInfo));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_START - IP : " IPSTR " , GW : " IPSTR " , NM : " IPSTR, __FILENAME__, __LINE__, IP2STR(&AdapterIPInfo.ip), IP2STR(&AdapterIPInfo.gw), IP2STR(&AdapterIPInfo.netmask));
        ESP_ERROR_CHECK(tcpip_adapter_get_dns_info(TCPIP_ADAPTER_IF_AP, TCPIP_ADAPTER_DNS_MAIN, &DnsInfo));
        inet_ntop(AF_INET, &DnsInfo.ip, ipString, sizeof(ipString));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_START - DNS - IP : %s", __FILENAME__, __LINE__, ipString);
        if (pIotModuleCom->hComServerTask != nullptr)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_START - xEventGroupSetBits", __FILENAME__, __LINE__);
          //  create the UDP socket and begin receive loop in ComServerTask
          xEventGroupSetBits(pIotModuleCom->hEventGroup, 1);
        }
        else
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - SYSTEM_EVENT_AP_START - ComServerTask not available", __FILENAME__, __LINE__);
        }
      }
      break;

    case SYSTEM_EVENT_AP_STOP:
      //  it must never happens
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - EVENT_AP_STOP", __FILENAME__, __LINE__);
      break;
    case SYSTEM_EVENT_AP_STACONNECTED:
      //  never seen
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STACONNECTED", __FILENAME__, __LINE__);
      break;
    case SYSTEM_EVENT_AP_STADISCONNECTED:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STADISCONNECTED", __FILENAME__, __LINE__);
      break;
    case SYSTEM_EVENT_AP_PROBEREQRECVED:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_PROBEREQRECVED", __FILENAME__, __LINE__);
      break;
    case SYSTEM_EVENT_AP_STAIPASSIGNED:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STAIPASSIGNED", __FILENAME__, __LINE__);
      //=====================================================================================
      //  //  dhcp_search_ip_on_mac does not compile, compiler throws "undefined reference to `dhcp_search_ip_on_mac(unsigned char*, ip4_addr*)'"
      //  {
      //    wifi_sta_list_t station_list;
      //    esp_wifi_ap_get_sta_list(&station_list);
      //    wifi_sta_info_t *stations = (wifi_sta_info_t*)station_list.sta;
      //    dhcp_search_ip_on_mac(stations[0].mac , &event->event_info.got_ip.ip_info.ip);
      //  }
      //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "WifiEventHandler - IPv4 : %s - changed : %d", ip4addr_ntoa(&event->event_info.got_ip.ip_info.ip), event->event_info.got_ip.ip_changed);
      //=====================================================================================
      break;
    case SYSTEM_EVENT_AP_STA_GOT_IP6:
      //  not implemented
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STA_GOT_IP6", __FILENAME__, __LINE__);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STA_GOT_IP6 - IPv6 : " IPV6STR, __FILENAME__, __LINE__, IPV62STR(event->event_info.got_ip6.ip6_info.ip));
      break;
    default:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_NOT_HANDLED - Event : %d", __FILENAME__, __LINE__, event->event_id);
      break;
  }
  //  feed event to arduino loop()
  if (pIotModuleCom->WifiEventHandlerCallback != nullptr) pIotModuleCom->WifiEventHandlerCallback(nullptr, event);
  return ret;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                   UDP Communication server task
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleCom::ComServerTask(void *pvParameters)
{
  struct sockaddr_in SockAddr;
  char pComBuf[UDP_COM_BUF_LEN];
  int ComBufLen;
#ifdef IOTMODULE_DEBUG
  int64_t Begin_uSec, End_uSec;
#endif
  bool StayInLoop = true;

  cIotModuleCom *pIotModuleCom = (cIotModuleCom *)pvParameters;

  do
  {
    //  start when station or soft AP initialization is done
    EventBits_t WaitBits = xEventGroupWaitBits(pIotModuleCom->hEventGroup, 1, pdTRUE, pdFALSE, portMAX_DELAY );
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - xEventGroupSetBits : %d", __FILENAME__, __LINE__, WaitBits);
    //  if at least one bit set the event is received else skip the code body because timeout elapsed (portMAX_DELAY means about 40 days) and loop again
    if (WaitBits != 0)
    {
      //  create UDP socket
      if ((pIotModuleCom->hIotHALSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) != -1)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - UDP socket created", __FILENAME__, __LINE__);

        //  not necessary
        //  SO_BROADCAST
        //  Set or get the broadcast flag. When enabled, datagram sockets are allowed to send packets to a broadcast address. This option has no effect on stream-oriented sockets.
        //  In this case we don't send broadcast, we only receive them so is not necessary
        //  int opt_flag = 1;
        //  ESP_ERROR_CHECK(setsockopt(hIotHALSocket, SOL_SOCKET, SO_BROADCAST, &opt_flag, sizeof(opt_flag)));

        SockAddr.sin_family = AF_INET;
        SockAddr.sin_port = htons(IOTMODULE_UDP_PORT_NUMBER);
        SockAddr.sin_addr.s_addr = htonl(INADDR_ANY);
        socklen_t SockAddrLen;

        //  set socket address length
        SockAddrLen = sizeof(struct sockaddr_in);
        // bind socket
        if (bind(pIotModuleCom->hIotHALSocket, (struct sockaddr *)&SockAddr, SockAddrLen) == 0)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - bind active, falling into recvfrom loop", __FILENAME__, __LINE__);
#ifdef IOTMODULE_DEBUG
          //  start perf measure
          Begin_uSec = esp_timer_get_time();
#endif
          do
          {
            //  get the last inputs before wait for new RPC
            pIotModuleCom->IotModuleApp.GetInputValues();
#ifdef IOTMODULE_DEBUG
            //  stop measure active phase
            End_uSec = esp_timer_get_time();
            //  set active phase value in generic
            pIotModuleCom->IotModuleApp.GenericWriteFloat(3, End_uSec - Begin_uSec);
            //  start measure inactive phase
            Begin_uSec = End_uSec;
#endif
            //  set socket address length
            SockAddrLen = sizeof(struct sockaddr_in);
            //  received 0 bytes (when used nmap -sU) or 1 or more bytes during normal operation
            if ((ComBufLen = recvfrom(pIotModuleCom->hIotHALSocket, pComBuf, UDP_COM_BUF_LEN, 0, (struct sockaddr *) &SockAddr, &SockAddrLen)) != -1)
            {
#ifdef IOTMODULE_DEBUG              
              //  stop measure inactive phase
              End_uSec = esp_timer_get_time();
              //  set inactive phase value in generic
              pIotModuleCom->IotModuleApp.GenericWriteFloat(9, End_uSec - Begin_uSec);
              //  start measure active phase
              Begin_uSec = End_uSec;
#endif
              //  reset watchdog
              pIotModuleCom->IO_WatchDogCounter = 0;

              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received packet from %s:%d", __FILENAME__, __LINE__, inet_ntoa(SockAddr.sin_addr), ntohs(SockAddr.sin_port));
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received data - len : %d", __FILENAME__, __LINE__, ComBufLen);

              //  RPC data, decode request and encode reply
              pIotModuleCom->NetModuleRPC((void*)pComBuf, &ComBufLen);

              //  if replay requested
              if (ComBufLen != 0)
              {
                //  if replay data can be transmitted
                if (ComBufLen <= sizeof(pComBuf))
                {
                  int Len = sendto(pIotModuleCom->hIotHALSocket, pComBuf, ComBufLen, 0, (struct sockaddr *)&SockAddr, SockAddrLen);
                  //  sent ok
                  if ((Len != -1) && (Len == ComBufLen))
                  {
                    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - sendto : sent packet to %s:%d", __FILENAME__, __LINE__, inet_ntoa(SockAddr.sin_addr), ntohs(SockAddr.sin_port));
                    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - sendto : data - len : %d", __FILENAME__, __LINE__, ComBufLen);

                    //  set the last ouput values received from RPC
                    pIotModuleCom->IotModuleApp.SetOutputValues(false);
                    StayInLoop = true;
                  }
                  //  ENOMEM
                  else if ((Len == -1) && (errno == ENOMEM))
                  {
                    //  https://docs.espressif.com/projects/esp-idf/en/latest/api-guides/lwip.html
                    //  Calling send() or sendto() repeatedly on a UDP socket may eventually fail with errno equal to ENOMEM. 
                    //  This is a limitation of buffer sizes in the lower layer network interface drivers. 
                    //  If all driver transmit buffers are full then UDP transmission will fail. 
                    //  Applications sending a high volume of UDP datagrams who don’t wish for any to be dropped by the sender should check for this error code and re-send the datagram after a short delay. 
                    //  Increasing the number of TX buffers in the Wi-Fi or Ethernet project configuration (as applicable) may also help.
                    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sendto ENOMEM error", __FILENAME__, __LINE__);
                    StayInLoop = true;
                  }
                  //  other error
                  else
                  {
                    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sendto error : %d", __FILENAME__, __LINE__, errno);
                    StayInLoop = false;
                  }
                }
                //  data size exeed buffer size
                else
                {
                  ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sendto - data size exeed buffer size", __FILENAME__, __LINE__);
                  StayInLoop = false;
                }
              }
              //  data not recognized, data size equal zero
              else
              {
                ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - sendto - 0 data size", __FILENAME__, __LINE__);
                StayInLoop = true;
              }
            }
            else
            //  recvfrom error
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - recvfrom error : %d", __FILENAME__, __LINE__, errno);
              StayInLoop = false;
            }
          }
          while (StayInLoop);
        }
        //  bind error
        else
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - socket bind error : %d", __FILENAME__, __LINE__, errno);
          StayInLoop = false;
        }
        close(pIotModuleCom->hIotHALSocket);
      }
      //  error creating socket
      else
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - UDP socket creation error : %d", __FILENAME__, __LINE__, errno);
        StayInLoop = false;
      }
    }
  }
  while (StayInLoop);

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - FATAL ERROR! communication task deleted!", __FILENAME__, __LINE__);
  TaskHandle_t hTmp = pIotModuleCom->hComServerTask;
  pIotModuleCom->hComServerTask = nullptr;
  vTaskDelete(hTmp);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                   Remote Procedure Call
//                  Set data structures to / from  communication data buffer
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleCom::NetModuleRPC(const void *pComBuf, int *pIotHALDataLen)
{
  tIotHALData *pIotHALData = (tIotHALData *)pComBuf;

  //  Received at least Protocol and ProtocolVer?
  if (*pIotHALDataLen >= (STRUCTMEMBERSIZE(sHeader, Protocol) + STRUCTMEMBERSIZE(sHeader, ProtocolVer)))
  {
    //  Protocol and ProtocolVer?
    if (memcmp(pIotHALData->Header.Protocol, IOTMODULE_THEREMINO_IOT_PROTOCOL, 6) == 0 && pIotHALData->Header.ProtocolVer == IOTMODULE_THEREMINO_IOT_PROTOCOL_VER)
    {
      //  FirmwareVer ignored for now
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.Who : %c", __FILENAME__, __LINE__, pIotHALData->Header.Who);
      //  Who?
      if (pIotHALData->Header.Who == 'H')
      {
        //  Command?
        switch (pIotHALData->Header.FuncId)
        {
          case SCAN:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : SCAN", __FILENAME__, __LINE__);

              //  Scan replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              memcpy(&pIotHALData->Body.ScanReply.MAC, WiFiMAC, sizeof(pIotHALData->Body.ScanReply.MAC));
              IotModuleApp.RPCGetModuleName(&pIotHALData->Body.ScanReply.ModuleNameSize, pIotHALData->Body.ScanReply.ModuleName);
              pIotHALData->Body.ScanReply.PinoutType = IOTMODULE_PINOUT_TYPE;
              pIotHALData->Body.ScanReply.ProgStatus = IotModuleApp.GetProgStatus();
              pIotHALData->Body.ScanReply.TotalPinsNumber = IotModuleApp.RPCGetTotalPinsNumber();
              for (uint8_t i = 0; i < IotModuleApp.RPCGetTotalPinsNumber(); i++)
              {
                pIotHALData->Body.ScanReply.PinName[i] = IotModuleApp.RPCGetPinName(i);
              }
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.ScanReply);
            }
            break;

          case PROG:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : PROG", __FILENAME__, __LINE__);

              //  Prog request RPC
              IotModuleApp.RPCSetAllPinsParams(pIotHALData->Body.ProgReq.TotalPinsNumber, pIotHALData->Body.ProgReq.PinType);

              //  Prog replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.ProgReply.ProgStatus = IotModuleApp.GetProgStatus();
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.ProgReply);
            }
            break;

          case GETSINGLE:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : GETSINGLE", __FILENAME__, __LINE__);

              //  GetPinTypes request RPC
              uint8_t tmp1 = pIotHALData->Body.GetSingleReq.PinIndex;

              //  GetPinTypes replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.GetSingleReply.PinName = IotModuleApp.RPCGetPinName(tmp1);
              IotModuleApp.RPCGetDynamicPinTypes(tmp1, &pIotHALData->Body.GetSingleReply.TotalPinTypes, pIotHALData->Body.GetSingleReply.PinType);
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.GetSingleReply);
            }
            break;

          case XCHG:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : XCHG", __FILENAME__, __LINE__);

              if (IotModuleApp.GetProgStatus() == 'Y')
              {
                //  Xchg request RPC (outputs)
                uint8_t tmp1 = pIotHALData->Body.XchgReq.TotalOutValues;
                if (IotModuleApp.GetOutputPinsNumber() == tmp1)
                {
                  for (uint8_t i = 0; i < tmp1; i++)
                  {
                    IotModuleApp.RPCSetOutputXchgValues(i, &pIotHALData->Body.XchgReq.OutValue[i]);
                  }

                  //  Xchg replay RPC (inputs)
                  //  pIotHALData->Header.Command = <the same>
                  pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
                  pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
                  pIotHALData->Header.Who = 'M';
                  //  speed it up, do not call GetProgrammed, test done before
                  pIotHALData->Body.XchgReply.ProgStatus = 'Y';
                  tmp1 = IotModuleApp.GetInputPinsNumber();
                  pIotHALData->Body.XchgReply.TotalInValues = tmp1;
                  for (uint8_t i = 0; i < tmp1; i++)
                  {
                    IotModuleApp.RPCGetInputXchgValues(i, &pIotHALData->Body.XchgReply.InValue[i]);
                  }
                  *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.XchgReply);
                }
                else
                {
                  ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Header.FuncId : XCHG - output pin number mismatch - req : %d - prog : %d", __FILENAME__, __LINE__, tmp1, IotModuleApp.GetOutputPinsNumber());

                  //  Xchg replay RPC (inputs)
                  //  pIotHALData->Header.Command = <the same>
                  pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
                  pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
                  pIotHALData->Header.Who = 'M';
                  //  todoo ProgStatus = 'N'
                  pIotHALData->Body.XchgReply.ProgStatus = 'N';
                  pIotHALData->Body.XchgReply.TotalInValues = 0;
                  *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.XchgReply);
                }
              }
              else
              {
                ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Header.FuncId : XCHG - called xchg but module not programmed", __FILENAME__, __LINE__);

                //  Xchg replay RPC (inputs)
                //  pIotHALData->Header.Command = <the same>
                pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
                pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
                pIotHALData->Header.Who = 'M';
                //  speed it up, do not call GetProgrammed, test done before
                pIotHALData->Body.XchgReply.ProgStatus = 'N';
                pIotHALData->Body.XchgReply.TotalInValues = 0;
                *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.XchgReply);
              }
            }
            break;

          case NAME:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : NAME", __FILENAME__, __LINE__);

              //  Name request RPC
              IotModuleApp.RPCSetModuleName(pIotHALData->Body.NameReq.ModuleNameSize, pIotHALData->Body.NameReq.ModuleName);

              //  Name replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              IotModuleApp.RPCGetModuleName(&pIotHALData->Body.NameReply.ModuleNameSize, pIotHALData->Body.NameReply.ModuleName);
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.NameReply);
            }
            break;

          case EXEC:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : EXEC", __FILENAME__, __LINE__);

              if (IOTMODULE_FUOTA_ENABLE == 1)
              {
                //  stop app dangerous activities
                IotModuleApp.PreFuota();
                //  start Firmware Upgrade On The Air
                Fuota();
              }
            }
            break;

          case PWM:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : PWM", __FILENAME__, __LINE__);

              //  Pwm request RPC
              uint8_t PinIndex = pIotHALData->Body.PwmReq.PinIndex;
              float Freq = pIotHALData->Body.PwmReq.Freq;
              IotModuleApp.RPCPwmSetFrequency(PinIndex, Freq);

              //  Pwm replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.PwmReply.ReceivedOK = true;
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.PwmReply);
            }
            break;
          //  todoo stepper
          case STEP:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : STEP", __FILENAME__, __LINE__);

              //  Step request RPC
              uint8_t PinIndex = pIotHALData->Body.StepReq.PinIndex;
              uint32_t SpeedHz = pIotHALData->Body.StepReq.SpeedHz;
              uint16_t AccHz_mSec = pIotHALData->Body.StepReq.AccHz_mSec;
              bool RetVal = IotModuleApp.StepperConfigSpeedAcc(PinIndex, SpeedHz, AccHz_mSec);
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - PinIndex : %d, SpeedHz : %d , AccHz_mSec : %d", __FILENAME__, __LINE__, PinIndex, SpeedHz, AccHz_mSec);
              
              //  Step replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.StepReply.ReceivedOK = RetVal;
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.StepReply);
            }
            break;
          case ADC24:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : ADC24", __FILENAME__, __LINE__);

              //  ADC24 request RPC
              uint16_t Sps = pIotHALData->Body.Adc24Req.Sps;
              bool MultipleSetups = pIotHALData->Body.Adc24Req.MultipleSetups;
              bool FiltersSingleCycle = pIotHALData->Body.Adc24Req.FiltersSingleCycle;
              uint8_t PreFilterType = pIotHALData->Body.Adc24Req.PreFilterType;
              uint8_t PostFilterType = pIotHALData->Body.Adc24Req.PostFilterType;
              uint16_t ActiveChannels = pIotHALData->Body.Adc24Req.ActiveChannels;
              uint16_t BiasedChannels = pIotHALData->Body.Adc24Req.BiasedChannels;
              uint16_t DiffChannels = pIotHALData->Body.Adc24Req.DiffChannels;
              uint16_t PseudoChannels = pIotHALData->Body.Adc24Req.PseudoChannels;
              uint8_t ChannelsGains0 = pIotHALData->Body.Adc24Req.ChannelsGains[0];
              uint8_t ChannelsGains1 = pIotHALData->Body.Adc24Req.ChannelsGains[1];
              uint8_t ChannelsGains2 = pIotHALData->Body.Adc24Req.ChannelsGains[2];
              IotModuleApp.ADC24SetParam(Sps, MultipleSetups, FiltersSingleCycle, PreFilterType,
                                         PostFilterType, ActiveChannels, BiasedChannels, DiffChannels,
                                         PseudoChannels, ChannelsGains0, ChannelsGains1, ChannelsGains2);
              //  ADC24 replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.Adc24Reply.ReceivedOK = true;
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.Adc24Reply);
            }
            break;
          default:
            //  FUNC_ID not recognized
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - FUNC_ID not recognized", __FILENAME__, __LINE__);
            //  do not replay
            *pIotHALDataLen = 0;
            break;
        }
      }
      else
      {
        //  sender is not HAL
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - sender is not HAL", __FILENAME__, __LINE__);
        //  do not replay
        *pIotHALDataLen = 0;
      }
    }
    else
    {
      //  protocol not recognized
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - protocol not recognized", __FILENAME__, __LINE__);
      //  do not replay
      *pIotHALDataLen = 0;
    }
  }
  else
  {
    //  data malformed
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - data malformed", __FILENAME__, __LINE__);
    //  do not replay
    *pIotHALDataLen = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                         Firmware Upgrade On The Air (FUOTA) fo Arduino IDE
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleCom::Fuota(void)
{
  //  MDNS TCP port
  int port = 3232;
  char hostName[32 + 1 + 12];
  mdns_txt_item_t arduTxtData[4] =
  {
    {(char*)"board"         , (char*)ARDUINO_VARIANT},
    {(char*)"tcp_check"     , (char*)"no"},
    {(char*)"ssh_upload"    , (char*)"no"},
    {(char*)"auth_upload"   , (char*)"no"}
  };

  char pModuleName[32 + 1];
  IotModuleApp.GetModuleName(pModuleName, sizeof(pModuleName));

  //  restart the module after 120 seconds if FUOTA is never done
  FUOTA_WatchDogEnable = true;

  sprintf(hostName, "%s-%02X:%02X:%02X:%02X:%02X:%02X", pModuleName, WiFiMAC[0], WiFiMAC[1], WiFiMAC[2], WiFiMAC[3], WiFiMAC[4], WiFiMAC[5]);
  ESP_ERROR_CHECK(mdns_init());
  ESP_ERROR_CHECK(mdns_hostname_set(hostName));
  ESP_ERROR_CHECK(mdns_service_add(NULL, "_arduino", "_tcp", port, arduTxtData, 4));

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - MDNS executed : %s", __FILENAME__, __LINE__, hostName);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - starting FUOTA", __FILENAME__, __LINE__);

  struct sockaddr_in ArduinoIDE_UdpAddr;
  char pArduinoIDE_UdpData[UDP_COM_BUF_LEN];
  unsigned int ArduinoIDE_UdpDataLen;
  int hArduinoIDE_UdpSocket;

  // create UDP socket
  if ((hArduinoIDE_UdpSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) >= 0)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - UDP socket created", __FILENAME__, __LINE__);

    //  SO_BROADCAST
    //  Set or get the broadcast flag. When enabled, datagram sockets are allowed to send packets to a broadcast address. This option has no effect on stream-oriented sockets.
    //  In this case we don't send broadcast, we only receive them so it is not necessary
    //  int opt_flag = 1;
    //  ESP_ERROR_CHECK(setsockopt(hArduinoIDE_UdpSocket, SOL_SOCKET, SO_BROADCAST, &opt_flag, sizeof(opt_flag)));

    ArduinoIDE_UdpAddr.sin_family = AF_INET;
    ArduinoIDE_UdpAddr.sin_port = htons(ARDUINO_IDE_UDP_PORT_NUMBER);
    ArduinoIDE_UdpAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    // bind socket
    if (bind(hArduinoIDE_UdpSocket, (struct sockaddr *)&ArduinoIDE_UdpAddr, sizeof(ArduinoIDE_UdpAddr)) >= 0)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - bind active, falling into recvfrom loop", __FILENAME__, __LINE__);

      if ((ArduinoIDE_UdpDataLen = recvfrom(hArduinoIDE_UdpSocket, pArduinoIDE_UdpData, UDP_COM_BUF_LEN, 0, (struct sockaddr *)&ArduinoIDE_UdpAddr, &ArduinoIDE_UdpDataLen)) <= 0)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - recvfrom error : %d", __FILENAME__, __LINE__, errno);
      }

      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received packet from %s:%d", __FILENAME__, __LINE__, inet_ntoa(ArduinoIDE_UdpAddr.sin_addr), ntohs(ArduinoIDE_UdpAddr.sin_port));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received raw data : %s - len %d", __FILENAME__, __LINE__ , pArduinoIDE_UdpData, ArduinoIDE_UdpDataLen);

      char delimiter = ' ';
      std::string tobesplit(pArduinoIDE_UdpData);
      std::vector<std::string> tokens;
      std::string token;
      std::istringstream tokenStream(tobesplit);
      while (std::getline(tokenStream, token, delimiter))
      {
        tokens.push_back(token);
      }

      if ((tokens.size() == 4) && (atoi(tokens[0].c_str()) == 0))
      {
        int OTA_TcpPort = atoi(tokens[1].c_str());
        int OTA_AppSize = atoi(tokens[2].c_str());
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received data - OTA_TcpPort : %d - OTA_AppSize %d - MD5 : %s", __FILENAME__, __LINE__,  OTA_TcpPort, OTA_AppSize, tokens[3].c_str());

        //  call esp_ota_begin here, before TCP initialization, to avoid strange behaviour in TCP session
        //  esp_ota_begin call "freeze" communications for about 3 seconds
        //  todo : bug? retry with newer arduino esp32 board's library version
        const esp_partition_t *update_partition = esp_ota_get_next_update_partition(NULL);
        esp_ota_handle_t update_handle;
        ESP_ERROR_CHECK(esp_ota_begin(update_partition, OTA_SIZE_UNKNOWN, &update_handle));

        sprintf(pArduinoIDE_UdpData, "%s", "OK");
        ArduinoIDE_UdpDataLen = strlen(pArduinoIDE_UdpData);
        if (sendto(hArduinoIDE_UdpSocket, pArduinoIDE_UdpData, ArduinoIDE_UdpDataLen, 0, (struct sockaddr *)&ArduinoIDE_UdpAddr, sizeof(ArduinoIDE_UdpAddr)) < 0)
        {
          ESP_ERROR_CHECK(esp_ota_end(update_handle));
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sendto error : %d", __FILENAME__, __LINE__, errno);
        }

        struct sockaddr_in ArduinoIDE_TcpAddr;
        int hArduinoIDE_TcpSocket;
        char pArduinoIDE_TcpData[FUOTA_TCP_DATA_BUF_LEN];
        unsigned int ArduinoIDE_TcpDataLen;

        //  create TCP socket
        if ((hArduinoIDE_TcpSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) >= 0)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TCP socket created", __FILENAME__, __LINE__);

          ArduinoIDE_TcpAddr.sin_addr.s_addr = ArduinoIDE_UdpAddr.sin_addr.s_addr;
          ArduinoIDE_TcpAddr.sin_family = AF_INET;
          ArduinoIDE_TcpAddr.sin_port = htons(OTA_TcpPort);
          if (connect(hArduinoIDE_TcpSocket, (struct sockaddr *)&ArduinoIDE_TcpAddr, sizeof(ArduinoIDE_TcpAddr)) == 0)
          {
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TCP socket connected, starting FUOTA", __FILENAME__, __LINE__);

            unsigned int OTA_AppSizeCounter = 0;
            unsigned int OTA_AppChunk;
            MD5Builder MD5;
            MD5.begin();

            do
            {
              ArduinoIDE_TcpDataLen = recv(hArduinoIDE_TcpSocket, pArduinoIDE_TcpData, FUOTA_TCP_DATA_BUF_LEN, 0);
              ESP_ERROR_CHECK(esp_ota_write(update_handle, (const void *)pArduinoIDE_TcpData, ArduinoIDE_TcpDataLen));
              OTA_AppSizeCounter += ArduinoIDE_TcpDataLen;
              OTA_AppChunk = ArduinoIDE_TcpDataLen;
              MD5.add(reinterpret_cast<uint8_t*>(pArduinoIDE_TcpData), ArduinoIDE_TcpDataLen);
              sprintf(pArduinoIDE_TcpData, "%u", ArduinoIDE_TcpDataLen);
              ArduinoIDE_TcpDataLen = strlen(pArduinoIDE_TcpData);
              ArduinoIDE_TcpDataLen = send(hArduinoIDE_TcpSocket, pArduinoIDE_TcpData, ArduinoIDE_TcpDataLen, 0);
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - chunk %u - total %u", __FILENAME__, __LINE__, OTA_AppChunk, OTA_AppSizeCounter);
            }
            while ((ArduinoIDE_TcpDataLen > 0) && (OTA_AppSize != OTA_AppSizeCounter));

            sprintf(pArduinoIDE_TcpData, "%s", "OK");
            ArduinoIDE_TcpDataLen = strlen(pArduinoIDE_TcpData);
            ArduinoIDE_TcpDataLen = send(hArduinoIDE_TcpSocket, pArduinoIDE_TcpData, ArduinoIDE_TcpDataLen, 0);
            close(hArduinoIDE_TcpSocket);
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TCP socket closed", __FILENAME__, __LINE__);

            ESP_ERROR_CHECK(esp_ota_end(update_handle));
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - esp_ota_end", __FILENAME__, __LINE__);

            MD5.calculate();
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - MD5s : %s", __FILENAME__, __LINE__, tokens[3].c_str());
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - MD5d : %s", __FILENAME__, __LINE__, MD5.toString().c_str());
            char pMD5s[33], pMD5d[33];
            unsigned int pMD5sl, pMD5dl;
            strncpy(pMD5s, tokens[3].c_str(), 32);
            pMD5s[32] = 0;
            strcpy(pMD5d, MD5.toString().c_str());
            pMD5sl = strlen(pMD5s);
            pMD5dl = strlen(pMD5d);
            int r = strcmp(pMD5s, pMD5d);
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - pMD5sl : %d - pMD5dl : %d - strcmp : %d", __FILENAME__, __LINE__, pMD5sl, pMD5dl, r);
            if (r == 0)
            {
              ESP_ERROR_CHECK(esp_ota_set_boot_partition(update_partition));
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - esp_ota_set_boot_partition", __FILENAME__, __LINE__);
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - FUOTA completed", __FILENAME__, __LINE__);
              //  todo : stop watchdog task
            }
          }
          else
          {
            ESP_ERROR_CHECK(esp_ota_end(update_handle));
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - TCP socket connection error : %d", __FILENAME__, __LINE__, errno);
          }
        }
        else
        {
          ESP_ERROR_CHECK(esp_ota_end(update_handle));
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - TCP socket creation error : %d", __FILENAME__, __LINE__, errno);
        }
      }
      close(hArduinoIDE_UdpSocket);
    }
    else
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - socket bind error : %d", __FILENAME__, __LINE__, errno);
      close(hArduinoIDE_UdpSocket);
      hArduinoIDE_UdpSocket = 0;
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - UDP socket creation error : %d", __FILENAME__, __LINE__, errno);
  }

  ESP_ERROR_CHECK(mdns_service_remove_all());
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - mdns_service_remove_all", __FILENAME__, __LINE__);
  //  this delay is needed between mdns_service_remove_all() and mdns_free()
  //  todo : bug? retry with newer arduino esp32 board's library version
  vTaskDelay(1000 / portTICK_PERIOD_MS);
  mdns_free();
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - mdns_free", __FILENAME__, __LINE__);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - esp_restart in 2 sec", __FILENAME__, __LINE__);
  vTaskDelay(2000 / portTICK_PERIOD_MS);
  esp_restart();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                  TimerCallback called every x mSec
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleCom::TimerCallback(TimerHandle_t pxTimer)
{
  cIotModuleCom *pIotModuleCom = (cIotModuleCom *)pvTimerGetTimerID( pxTimer );

  if (IOTMODULE_OUTPUT_WATCHDOG_ENABLE == 1)
  {
    //  if delay not triggered yet
    if (pIotModuleCom->IO_WatchDogCounter <= IO_WATCHDOG_DELAY_MS / TIMER_MS)
    {
      //  if delay triggered
      if (pIotModuleCom->IO_WatchDogCounter == IO_WATCHDOG_DELAY_MS / TIMER_MS)
      {
        //  reset output values
        pIotModuleCom->IotModuleApp.SetOutputValues(true);
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - output watchdog triggered, output values have been reset", __FILENAME__, __LINE__);
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - TimerCallback execution core : %d", __FILENAME__, __LINE__, xPortGetCoreID());
      }
      pIotModuleCom->IO_WatchDogCounter++;
    }
  }
  if (pIotModuleCom->FUOTA_WatchDogEnable)
  {
    //  if delay not triggered yet
    if (pIotModuleCom->FUOTA_WatchDogCounter <= FUOTA_WATCHDOG_DELAY_MS / TIMER_MS)
    {
      //  if delay triggered
      if (pIotModuleCom->FUOTA_WatchDogCounter == FUOTA_WATCHDOG_DELAY_MS / TIMER_MS)
      {
        //  log reboot
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - esp_restart in 2 sec", __FILENAME__, __LINE__);
        //  delay 2 Sec
        vTaskDelay(2000 / portTICK_PERIOD_MS);
        //  reboot system
        esp_restart();
      }
      pIotModuleCom->FUOTA_WatchDogCounter++;
    }
  }
  //  Stepper interpolator
  pIotModuleCom->IotModuleApp.TimedInterpolate();
}
