class cIotModuleApp;

//================================================================================================
//                                  cIotModuleServo header
//================================================================================================

class cIotModuleServo
{
  private:

    static const uint8_t MAX_CHANNELS = 16;
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_20_BIT;
    static const uint32_t SPEED = 50;

    struct sSingleParams
    {
      bool Attached;
    };

    struct sAllData
    {
      struct sSingleParams Params[MAX_CHANNELS];
    };

    struct sAllData Data;
    cIotModuleApp *pIotModuleApp;
    ledc_timer_config_t ledc_timer;
    ledc_channel_config_t ledc_channel;

  public:

    void Initialize(cIotModuleApp *pIotModuleApp);
    bool Attach(const uint8_t LedcChannel, const uint8_t LedcTimer, const gpio_num_t Name);
    bool Detach(const uint8_t LedcChannel);
    bool SetDuty(const uint8_t LedcChannel, float Duty);
};

//================================================================================================
//                                  cIotModulePWM header
//================================================================================================

class cIotModulePWM
{
  private:

    static const uint8_t MAX_CHANNELS = 16;
    static const ledc_timer_bit_t LEDC_DUTY = LEDC_TIMER_16_BIT;
    static const uint32_t SPEED = 50;

    struct sSingleParams
    {
      gpio_num_t Name;
      uint32_t MaxWidth;
      bool Attached;
    };

    struct sAllData
    {
      struct sSingleParams Params[MAX_CHANNELS];
    };

    struct sAllData Data;
    cIotModuleApp *pIotModuleApp;
    ledc_timer_config_t ledc_timer;
    ledc_channel_config_t ledc_channel;
    
  public:

    void Initialize(cIotModuleApp *pIotModuleApp);
    bool Attach(const uint8_t LedcChannel, const uint8_t LedcTimer, const gpio_num_t Name);
    bool Detach(const uint8_t LedcChannel);
    bool SetDuty(const uint8_t LedcChannel, float Duty);
    bool SetFreq(const uint8_t LedcChannel, float Freq);
};

//================================================================================================
//                                  cIotModuleCounter header
//================================================================================================

class cIotModuleCounter
{
  public:

    static const uint8_t MAX_CHANNELS = 8;

  private:

    struct sSingleParams
    {
      gpio_num_t Name;
      gpio_pull_mode_t PullMode; 
      gpio_pull_mode_t ResetMode;
      bool Attached;
    };

    struct sAllData
    {
      struct sSingleParams Params[MAX_CHANNELS];
    };

    struct sAllData Data;
    pcnt_config_t pcnt_unit;
    cIotModuleApp *pIotModuleApp;

  public:

    void Initialize(cIotModuleApp *pIotModuleApp);
    void SetControlPin(const uint8_t Channel, const gpio_num_t Name, gpio_pull_mode_t PullMode, const gpio_pull_mode_t ResetMode);
    bool Attach(const uint8_t Channel);
    bool Detach(const uint8_t Channel);
    uint16_t GetCount(const uint8_t Channel);
};

//================================================================================================
//                                  cIotModuleHWEncoder header
//================================================================================================

class cIotModuleHWEncoder
{
  public:

    static const uint8_t MAX_CHANNELS = 8;
    static const uint8_t PIN_A = 0;
    static const uint8_t PIN_B = 1;

  private:

    struct sSingleParams
    {
      gpio_num_t PinA_Name;
      gpio_num_t PinB_Name;
      gpio_pull_mode_t PinA_PullMode; 
      gpio_pull_mode_t PinB_PullMode;
      gpio_pull_mode_t PinA_ResetMode;
      gpio_pull_mode_t PinB_ResetMode;
      bool Attached;
    };

    struct sAllData
    {
      struct sSingleParams Params[MAX_CHANNELS];
    };

    struct sAllData Data;
    pcnt_config_t pcnt_unit;
    cIotModuleApp *pIotModuleApp;

  public:

    void Initialize(cIotModuleApp *pIotModuleApp);
    void SetControlPin(const uint8_t Channel, uint8_t Type, const gpio_num_t Name, gpio_pull_mode_t PullMode, const gpio_pull_mode_t ResetMode);
    bool Attach(const uint8_t Channel);
    bool Detach(const uint8_t Channel);
    uint16_t GetCount(const uint8_t Channel);
};

//================================================================================================
//                                  cIotModuleSWEncoder header
//================================================================================================

class cIotModuleSWEncoder
{
  public:

    static const uint8_t MAX_CHANNELS = 8;

  private:

    const int8_t QuadratureTable[16] = {0, -1, 1, 0, 1, 0, 0, -1, -1, 0, 0, 1, 0, 1, -1, 0};

    struct sSingleParams
    {
      gpio_num_t PinA_Name;
      gpio_num_t PinB_Name;
      uint8_t PinAB_Status;
      uint8_t PinAB_StatusOld;
      uint16_t Count;
      bool Attached;
    };

    struct sAllData
    {
      struct sSingleParams Params[MAX_CHANNELS];
    };

    struct sAllData Data;

    static void OnChangeCommon(void *arg, const uint8_t Channel);
    static void ISR_OnChange_0(void *arg);
    static void ISR_OnChange_1(void *arg);
    static void ISR_OnChange_2(void *arg);
    static void ISR_OnChange_3(void *arg);
    static void ISR_OnChange_4(void *arg);
    static void ISR_OnChange_5(void *arg);
    static void ISR_OnChange_6(void *arg);
    static void ISR_OnChange_7(void *arg);

  public:

    void Initialize(cIotModuleApp *pIotModuleApp);
    void PreFuota(void);    
    bool Attach(const uint8_t Channel, gpio_num_t PinA_Name, gpio_num_t PinB_Name);
    bool Detach(const uint8_t Channel);
    bool GetCount(const uint8_t Channel, int16_t *Count);
};

//================================================================================================
//                                  cIotModulePeriod header
//================================================================================================

class cIotModulePeriod
{
  public:

    static const uint8_t MAX_CHANNELS = 8;

  private:

    struct sSingleParams
    {
      uint8_t PinName;
      uint32_t Count;
      uint32_t TimeCounterOld;
      bool Attached;
    };

    struct sAllData
    {
      struct sSingleParams Params[MAX_CHANNELS];
    };

    struct sAllData Data;

    static void OnRiseCommon(void *arg, const uint8_t Channel);
    static void ISR_OnRise_0(void *arg);
    static void ISR_OnRise_1(void *arg);
    static void ISR_OnRise_2(void *arg);
    static void ISR_OnRise_3(void *arg);
    static void ISR_OnRise_4(void *arg);
    static void ISR_OnRise_5(void *arg);
    static void ISR_OnRise_6(void *arg);
    static void ISR_OnRise_7(void *arg);

  public:

    void Initialize(cIotModuleApp *pIotModuleApp);
    void PreFuota(void);
    bool Attach(const uint8_t Channel, const uint8_t PinName);
    bool Detach(const uint8_t Channel);
    bool GetCount(const uint8_t Channel, uint32_t *Count);
};

//================================================================================================
//                                  cExtend_16_64 class
//================================================================================================

class cExtend_16_64
{
  private:
  
    static const int64_t Limit1 = 16384;
    static const int64_t Limit2 = 49152;
    static const int64_t Limit3 = 65536;

    int32_t mNewValue;
    int32_t mOldValue;
    int32_t mValue;

  public:
    
    int64_t Reset(const int16_t Value)
    {
      mOldValue = (int32_t)Value + 32768;
      mValue = 0;
      return mValue;
    }
    
    int64_t Set(const int32_t Value)
    {
      mValue = Value;
      return mValue;
    }

    int64_t Update(const int16_t Value)
    {
      mNewValue = (int32_t)Value + 32768;
      if (mNewValue != mOldValue)
      {
        mValue += mNewValue - mOldValue;
        if (mNewValue > Limit2 && mOldValue < Limit1) mValue -= Limit3;
        if (mNewValue < Limit1 && mOldValue > Limit2) mValue += Limit3;
        mOldValue = mNewValue;
      }
      return mValue;
    }
};

//================================================================================================
//                                  cIotModuleRes class
//================================================================================================

//  The resource calculator allows the validation or not of a configuration for a certain number of steppers and / or servos and / or pwm and / or counters and / or encoders.
//  If the configuration is valid, it lets you know if and which pin types you can still add the 5 types mentioned above.
//  The resources that are used for those 5 types of pins are represented inside: pcnt_unit (8) and ledc_unit (16); the latter are implicitly linked to the relative timers (8).
//  To facilitate its calculation, it groups the pins configured by type categories with the following descending priority sequence: stepper, servo, pwm, counter, encoder.
//  To correctly program the hw components, a function then distributes the identifiers of the necessary resources according to the configured order of the pins

//  SX column : servo/pwm channel 0...15
//  ledc_channel.speed_mode = Channel / 8 ? LEDC_LOW_SPEED_MODE(0) : LEDC_HIGH_SPEED_MODE(1);
//  ledc_channel.channel    = (ledc_channel_t)(Channel % 8);
//  ledc_channel.timer_sel  = (ledc_timer_t)((Channel / 2) % 4);

//  DX column : stepper 0...7
//  ledc_channel.speed_mode = Channel / 4 ? LEDC_LOW_SPEED_MODE(0) : LEDC_HIGH_SPEED_MODE(1);
//  ledc_channel.channel    = (ledc_channel_t)((Channel * 2) % 8);
//  ledc_channel.timer_sel  = (ledc_timer_t)(Channel % 4);

//    id          speed_mode         channel          timer_sel
//  0  | 0           1 | 1            0 | 0             0 | 0
//  1  |             1 |              1 |               0 |
//  2  | 1           1 | 1            2 | 2             1 | 1
//  3  |             1 |              3 |               1 |
//  4  | 2           1 | 1            4 | 4             2 | 2
//  5  |             1 |              5 |               2 |
//  6  | 3           1 | 1            6 | 6             3 | 3
//  7  |             1 |              7 |               3 |
//  8  | 4           0 | 0            0 | 0             0 | 0
//  9  |             0 |              1 |               0 |
//  10 | 5           0 | 0            2 | 2             1 | 1
//  11 |             0 |              3 |               1 |
//  12 | 6           0 | 0            4 | 4             2 | 2
//  13 |             0 |              5 |               2 |
//  14 | 7           0 | 0            6 | 6             3 | 3
//  15 |             0 |              7 |               3 |

//  timer_sel : 0...7
//  ledc_timer.speed_mode      = Channel / 4 ? LEDC_LOW_SPEED_MODE(0) : LEDC_HIGH_SPEED_MODE(1);
//  ledc_timer.timer_num       = (ledc_timer_t)(Channel % 4);

//  timer_sel    speed_mode    timer_num
//      0            1             0
//      1            1             1
//      2            1             2
//      3            1             3
//      4            0             0
//      5            0             1
//      6            0             2
//      7            0             3

class cIotModuleRes
{
  private:
  
    enum RESTYPE
    {
      FREE, //  Free
      UNAV, //  Unavailable
      ST,   //  Stepper
      SE,   //  Servo
      PW,   //  PWM
      CN,   //  Counter
      EN,   //  Encoder
      BUSY, //  Resurce allocated
    };

    const uint8_t MAX_ST = 8;
    const uint8_t MAX_SE = 16;
    const uint8_t MAX_P = 16;
    const uint8_t MAX_C = 8;

    uint8_t LedcST;
    uint8_t LedcSE;
    uint8_t LedcPW;
    uint8_t LedcFR;
    uint8_t Pcnt;

    RESTYPE LedcArray[16];
    RESTYPE PcntArray[8];

  public:

    cIotModuleRes()
    {
      ESP_ERROR_CHECK(ledc_fade_func_install(0));
    }
    
    void Initialize(void)
    {
      for (uint8_t i = 0; i < 16; i++)
      {
        LedcArray[i] = FREE;
      }
      for (uint8_t i = 0; i < 8; i++)
      {
        PcntArray[i] = FREE;
      }
      LedcST = 0;
      LedcSE = 0;
      LedcPW = 0;
      LedcFR = 0;
      Pcnt = 0;
    }

    bool Fill(uint8_t NumST, uint8_t NumSE, uint8_t NumPW, uint8_t NumCN, uint8_t NumEN)
    {
      bool RetVal = true;
      int8_t j, k;

      Initialize();

      j = k = 0;
      //  ledc for Stepper
      for (uint8_t i = 0; i < NumST && RetVal; i++)
      {
        if (j == 16)
        {
          RetVal = false;
          break;
        }
        LedcArray[j++ % 16] = ST;
        LedcArray[j++ % 16] = UNAV;

        if (k == 8)
        {
          RetVal = false;
          break;
        }
        PcntArray[k++ % 8] = ST;
      }
      //  ledc for Servo
      for (uint8_t i = 0; i < NumSE && RetVal; i++)
      {
        if (j == 16)
        {
          RetVal = false;
          break;
        }
        LedcArray[j++ % 16] = SE;
      }
      //  ledc for Servo odd
      if (RetVal == true)
      {
        if (j % 2 == 1) LedcArray[j++ % 16] = UNAV;
      }
      //  ledc for PWM
      uint8_t _NumPW = NumPW;
      //  forward
      for (uint8_t i = 0; i < NumPW && RetVal; i++)
      {
        if (j == 16)
        {
          j--;
          break;
        }
        LedcArray[j % 16] = PW;

        j+=2;
        _NumPW--;
      }
      //  reverse
      for (uint8_t i = 0; i < _NumPW && RetVal; i++)
      {
        if (LedcArray[j % 16] != FREE)
        {
          RetVal = false;
          break;
        }

        LedcArray[j % 16] = PW;
        if (j < 2)
        {
          RetVal = false;
          break;
        }
        j-=2;
      }
      //  pcnt for counters
      for (uint8_t i = 0; i < NumCN && RetVal; i++)
      {
        if (k == 8)
        {
          RetVal = false;
          break;
        }
        PcntArray[k++ % 8] = CN;
      }
      //  pcnt for EncoderHW
      for (uint8_t i = 0; i < NumEN && RetVal; i++)
      {
        if (k == 8)
        {
          RetVal = false;
          break;
        }
        PcntArray[k++ % 8] = EN;
      }
      //  used 1edc
      for (uint8_t i = 0; i < 16; i++)
      {
        switch (LedcArray[i])
        {
          case ST:
            LedcST++;
            break;
          case SE:
            LedcSE++;
            break;
          case PW:
            LedcPW++;
            break;
          case FREE:
            LedcFR++;
            break;
          default:
            break;
        }
      }
      //  used pcnt
      for (uint8_t i = 0; i < 8; i++)
      {
        switch (PcntArray[i])
        {
          case ST:
          case CN:
          case EN:
            Pcnt++;
            break;
          default:
            break;
        }
      }
      //  PrintLedcAlloc();
      //  PrintPcntAlloc();
      //  PrintAvail();
      if (!RetVal)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Resources allocation fault", __FILENAME__, __LINE__);
      }
      return RetVal;
    }

    bool AllocStepper(uint8_t *pLedcTimer, uint8_t *pLedcChannel, uint8_t *pPcntChannel)
    {
      bool RetVal1 = false, RetVal2 = false;
      for (uint8_t i = 0; i < 16; i+= 2)
      {
        if (LedcArray[i] == ST)
        {
          *pLedcChannel = i;
          *pLedcTimer = i / 2;
          LedcArray[i] = BUSY;
          RetVal1 = true;
          break;
        }
      }
      for (uint8_t i = 0; i < 8; i++)
      {
        if (PcntArray[i] == ST)
        {
          *pPcntChannel = i;
          PcntArray[i] = BUSY;
          RetVal2 = true;
          break;
        }
      }
      return RetVal1 && RetVal2;
    }

    bool AllocServo(uint8_t *pLedcChannel, uint8_t *pLedcTimer)
    {
      bool RetVal = false;
      for (uint8_t i = 0; i < 16; i++)
      {
        if (LedcArray[i] == SE)
        {
          *pLedcChannel = i;
          *pLedcTimer = i / 2;
          LedcArray[i] = BUSY;
          RetVal = true;
          break;
        }
      }
      return RetVal;
    }

    bool AllocPWM(uint8_t *pLedcChannel, uint8_t *pLedcTimer)
    {
      bool RetVal = false;
      for (uint8_t i = 0; i < 16; i++)
      {
        if (LedcArray[i] == PW)
        {
          *pLedcChannel = i;
          *pLedcTimer = i / 2;
          LedcArray[i] = BUSY;
          RetVal = true;
          break;
        }
      }
      return RetVal;
    }

    bool AllocCounter(uint8_t *pPcntChannel)
    {
      bool RetVal = false;
      for (uint8_t i = 0; i < 8; i++)
      {
        if (PcntArray[i] == CN)
        {
          *pPcntChannel = i;
          PcntArray[i] = BUSY;
          RetVal = true;
          break;
        }
      }
      return RetVal;
    }

    bool AllocHWEncoder(uint8_t *pPcntChannel)
    {
      bool RetVal = false;
      for (uint8_t i = 0; i < 8; i++)
      {
        if (PcntArray[i] == EN)
        {
          *pPcntChannel = i;
          PcntArray[i] = BUSY;
          RetVal = true;
          break;
        }
      }
      return RetVal;
    }

    bool MoreStepper()
    {
      //  limit stepper number to pcnt free
      uint8_t AvailST = LedcFR / 2;
      if(AvailST > MAX_C - Pcnt)
      {
        AvailST = MAX_C - Pcnt;
      }
      return AvailST != 0;
    }

    bool MoreServo()
    {
      return ((LedcFR / 2) * 2  + LedcSE % 2) != 0;
    }

    bool MorePWM()
    {
      return LedcFR != 0;
    }

    bool MoreCounter()
    {
      return  (MAX_C - Pcnt) != 0;
    }

    bool MoreHWEncoder()
    {
      return  (MAX_C - Pcnt) != 0;
    }

    void PrintLedcAlloc(void)
    {
      for (uint8_t i = 0; i < 16; i++)
      {
        switch (LedcArray[i])
        {
        case FREE:
          printf("FREE ");
          break;
        case UNAV:
          printf("UNAV ");
          break;
        case ST:
          printf(" ST  ");
          break;
        case SE:
          printf(" SE  ");
          break;
        case PW:
          printf(" PW  ");
          break;
        case BUSY:
          printf("BUSY ");
          break;
        default:
          break;
        }
        if (i % 2 == 1 && i != 15) printf("- ");
      }
      printf("\n");
    }

    void PrintPcntAlloc(void)
    {
      for (uint8_t i = 0; i < 8; i++)
      {
        switch (PcntArray[i])
        {
        case FREE:
          printf("FREE ");
          break;
        case ST:
          printf(" ST  ");
          break;
        case CN:
          printf(" CN  ");
          break;
        case EN:
          printf(" EN  ");
          break;
        case BUSY:
          printf("BUSY ");
          break;
        default:
          break;
        }
        if (i != 7) printf("- ");
      }
      printf("\n");
    }
    
    void PrintAvail(void)
    {
      printf("Available Stepper : %s\n", MoreStepper() ? "true" : "false");
      printf("Available Servo : %s\n", MoreServo() ? "true" : "false");
      printf("Available PWM : %s\n", MorePWM() ? "true" : "false");
      printf("Available Counter : %s\n", MoreCounter() ? "true" : "false");
      printf("Available EncoderHW : %s\n", MoreHWEncoder() ? "true" : "false");
    }

};

